/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <stddef.h>
#include "meta-comet.h"
#include "int-list.h"
#include <treedb/comet-guids.h>

CometCommandStatus int_list_setup
( meta_comet_ptr_t runtime
) {
	treedb_const_list_ptr_t plist_template
		= treedb_get_list_template_for
			(runtime->m_storage.m_pTreedb->m_uBytesPerOffset
			, eTreedbContainerHasCount);
	if(!plist_template) {
		fprintf(stderr, "No matching list template found.\n");
		return 1;
	}
	treedb_list_backend_t_Setup
		( & runtime->u_list.m_backend
		, & IntList_backend_vtbl
		);
	treedb_list_impl_t_Setup
		( & runtime->u_list.m_list_impl
		, & runtime->m_allocator
		, & runtime->u_list.m_backend
		, & runtime->u_list.m_description
		, plist_template->m_pImpl
		);
	/* We need to align the int32_t on a 4-byte boundary, like the compiler
		does, to be compatible with treedb, and as the backend expects it.
	*/
	uint16_t nsz
		= V3C_ROUND_UP_T2(uint16_t
			, runtime->u_list.m_description.m_node.m_uSize
			, sizeof(int32_t)
			);
	META_TREEDB_printf2("Changing aligned_node_size from %zu to %u.\n"
		, (size_t)runtime->u_list.m_description.m_aligned_node_size
		, nsz);
	runtime->u_list.m_description.m_aligned_node_size = nsz;

	treedb_list_t_Setup
		( & runtime->u_list.m_list
		, & runtime->u_list.m_list_impl
		, runtime->m_storage.m_pTreedb->m_ulPayloadOffset
		);
	return 0;
}
CometCommandStatus int_list_init
( meta_comet_ptr_t runtime
) {
	if(int_list_setup(runtime))
		return eCometCommandHeapError;
	treedb_list_t_Init(& runtime->u_list.m_list);
	return eCometCommandSuccess;
}

CometCommandStatus int_list_dump
( meta_comet_ptr_t runtime
) {
	if(int_list_setup(runtime))
		return eCometCommandHeapError;
	/*L2IntList_Dump(& runtime->u_list.m_list);*/
	treedb_list_ptr_t plist = & runtime->u_list.m_list;
	PVOID node = treedb_list_t_PointerHead(plist);
	for(; node; node = treedb_list_t_PointerNodeNextPtr(plist, node)) {
		int32_t * p = IntList_DataPointer(plist, node);
		META_TREEDB_printf2("[%zu] "
			, (size_t)treedb_tree_t_PointerToOffset(plist, p));
		printf("%d\n", (int)(*p));
	}
	return eCometCommandSuccess;
}
CometCommandStatus int_list_count
( meta_comet_ptr_t runtime
) {
	if(int_list_setup(runtime))
		return eCometCommandHeapError;
	size_t count = treedb_list_t_Count(& runtime->u_list.m_list);
	printf("%zu\n", count);
	return eCometCommandSuccess;
}
CometCommandStatus int_list_push_back
( meta_comet_ptr_t runtime
, const char * szitem
) {
	char * tailptr;
	int item = (int)strtol(szitem, & tailptr, 10);
	if(!tailptr || *tailptr) {
		  fprintf(stderr, "Error \"%s\" isn't a number.\n", szitem);
		  return eCometCommandBadCommand;
	}
	if(int_list_setup(runtime))
		return eCometCommandHeapError;

	TreedbHeapStatus res = treedb_storage_t_Unlock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = treedb_storage_t_WriteLock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	HTREEDB_LISTNODE res2 = IntList_PushBack
		(& runtime->u_list.m_list, item);
	if(!res2)
		return eCometCommandOutOfResources;
	return eCometCommandSuccess;
}
CometCommandStatus int_list_push_front
( meta_comet_ptr_t runtime
, const char * szitem
) {
	char * tailptr;
	int item = (int)strtol(szitem, & tailptr, 10);
	if(!tailptr || *tailptr) {
		  fprintf(stderr, "Error \"%s\" isn't a number.\n", szitem);
		  return eCometCommandBadCommand;
	}
	if(int_list_setup(runtime))
		return eCometCommandHeapError;

	TreedbHeapStatus res = treedb_storage_t_Unlock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = treedb_storage_t_WriteLock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	HTREEDB_LISTNODE res2 = IntList_PushFront
		(& runtime->u_list.m_list, item);
	if(!res2)
		return eCometCommandOutOfResources;
	return eCometCommandSuccess;
}
CometCommandStatus int_list_pop_back
( meta_comet_ptr_t runtime
) {
	if(int_list_setup(runtime))
		return eCometCommandHeapError;

	TreedbHeapStatus res = treedb_storage_t_Unlock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = treedb_storage_t_WriteLock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	BOOL ret = IntList_PopBack(& runtime->u_list.m_list);
	return ret == 0 ? eCometCommandFailure : eCometCommandSuccess;
}
CometCommandStatus int_list_pop_front
( meta_comet_ptr_t runtime
) {
	if(int_list_setup(runtime))
		return eCometCommandHeapError;

	TreedbHeapStatus res = treedb_storage_t_Unlock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = treedb_storage_t_WriteLock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	BOOL ret = IntList_PopFront(& runtime->u_list.m_list);
	return ret == 0 ? eCometCommandFailure : eCometCommandSuccess;
}
CometCommandStatus int_list_stream
( meta_comet_ptr_t runtime
) {
	if(int_list_setup(runtime))
		return eCometCommandHeapError;
	TreedbHeapStatus res = treedb_storage_t_Unlock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = treedb_storage_t_WriteLock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	int32_t n;
	CometCommandStatus ret = eCometCommandSuccess;
	int c;
	for(;;) {
		c = scanf("%d", &n);
		if(c != 1)
			break;
		ret = (IntList_PushBack(& runtime->u_list.m_list, n)
				 == 0)
			? eCometCommandOutOfResources : eCometCommandSuccess;
		if(ret == eCometCommandOutOfResources)
			break;
	}
	return ret;
}
CometCommandStatus int_list_type
( meta_comet_ptr_t runtime
, const char * filename
) {
	if(int_list_setup(runtime))
		return eCometCommandHeapError;
	/*int bad = treedb_allocator_t_Check(& runtime->m_allocator);*/
	int bad = 0;
	char u[37];
	uuid_unparse_lower(runtime->m_storage.m_pTreedb->m_uSchema, u);
	if(bad) {
		printf(
			"%s is a %u bit int list treedb [%s] [check failed - %d].\n"
			, filename, runtime->m_storage.m_pTreedb->m_uBytesPerOffset*8, u, bad);
	}else{
		size_t count = treedb_list_t_Count(& runtime->u_list.m_list);
		printf(
			"%s is a %u bit int list treedb [%s] with %zu entries.\n"
			, filename, runtime->m_storage.m_pTreedb->m_uBytesPerOffset*8, u
			, count);
	}
	fflush(stdout);
	treedb_allocator_t_DumpSchema(& runtime->m_allocator, STDOUT_FILENO);
	return eCometCommandSuccess;
}

CometDetails int_list_details =
{ UUID_treedb_int_list_schema
, & int_list_init
, 3
};

CometCommandStatus int_list_process
( meta_comet_ptr_t runtime
, int argc, char *argv[]
) {
	switch(argc) {
	case 5:
		/* prg set string <op> <filename> */
		if(!strcmp(argv[3], "pop_back"))
			return int_list_pop_back(runtime);
		if(!strcmp(argv[3], "pop_front"))
			return int_list_pop_front(runtime);
		break;
	case 6:
		/* prg set string <op> <filename> arg */
		if(!strcmp(argv[3], "push_back"))
			return int_list_push_back(runtime, argv[5]);
		if(!strcmp(argv[3], "push_front"))
			return int_list_push_front(runtime, argv[5]);
		break;
	}
	return eCometCommandUnrecognisedCommand;
}
