/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <stddef.h>
#include "meta-comet.h"
#include "int-tree.h"
#include <v3c/meta-avl.h> /* We need to start from somewhere. */

CometCommandStatus int_tree_setup
( meta_comet_ptr_t runtime
) {
	treedb_const_tree_ptr_t ptree_template
		= treedb_get_avl_tree_template_for
			(runtime->m_storage.m_pTreedb->m_uBytesPerOffset
			/* Map meta-treedb enum values to treedb ones : multiply by four. */
			, runtime->m_storage.m_pTreedb->m_uFeatures * 4);
	if(!ptree_template) {
		fprintf(stderr, "No matching tree template found.\n");
		return 1;
	}
	treedb_tree_backend_t_Setup
		( & runtime->u_tree.m_backend
		, & IntTree_backend_vtbl
		);
	treedb_tree_impl_t_Setup2
		( & runtime->u_tree.m_tree_impl
		, & runtime->m_allocator
		, & runtime->u_tree.m_backend
		, & runtime->u_tree.m_description
		, ptree_template->m_pImpl
		);
	/* We need to align the int32_t on a 4-byte boundary, like the compiler
		does, to be compatible with treedb, and as the backend expects it.
	*/
	uint16_t nsz
		= V3C_ROUND_UP_T2(uint16_t
			, runtime->u_tree.m_description.m_node.m_uSize
			, sizeof(int32_t)
			);
	META_TREEDB_printf2("Changing aligned_node_size from %zu to %u.\n"
		, (size_t)runtime->u_tree.m_description.m_aligned_node_size
		, nsz);
	runtime->u_tree.m_description.m_aligned_node_size = nsz;

	treedb_tree_t_Setup
		( & runtime->u_tree.m_tree
		, & runtime->u_tree.m_tree_impl
		, runtime->m_storage.m_pTreedb->m_ulPayloadOffset
		);
	return 0;
}
CometCommandStatus int_tree_init
( meta_comet_ptr_t runtime
) {
	if(int_tree_setup(runtime))
		return eCometCommandHeapError;
	treedb_tree_t_Init(& runtime->u_tree.m_tree);
	return eCometCommandSuccess;
}

CometCommandStatus int_tree_dump
( meta_comet_ptr_t runtime
) {
	if(int_tree_setup(runtime))
		return eCometCommandHeapError;
	/*IntTree_Dump(& runtime->u_tree.m_tree);*/
	treedb_tree_ptr_t ptree = & runtime->u_tree.m_tree;
	PVOID node = treedb_tree_t_PointerHead(ptree);
	for(; node; node = treedb_tree_t_PointerNodeNextPtr(ptree, node)) {
		int32_t * p = IntTree_DataPointer(ptree, node);
		META_TREEDB_printf2("[%zu] "
			, (size_t)treedb_tree_t_PointerToOffset(ptree, p));
		printf("%d\n", (int)(*p));
	}
	return eCometCommandSuccess;
}
CometCommandStatus int_tree_count
( meta_comet_ptr_t runtime
) {
	if(int_tree_setup(runtime))
		return eCometCommandHeapError;
	size_t count = treedb_tree_t_Count(& runtime->u_tree.m_tree);
	printf("%zu\n", count);
	return eCometCommandSuccess;
}
CometCommandStatus int_tree_insert
( meta_comet_ptr_t runtime
, const char * szitem
) {
	char * tailptr;
	int item = (int)strtol(szitem, & tailptr, 10);
	if(!tailptr || *tailptr) {
		  fprintf(stderr, "Error \"%s\" isn't a number.\n", szitem);
		  return eCometCommandBadCommand;
	}
	if(int_tree_setup(runtime))
		return eCometCommandHeapError;

	TreedbHeapStatus res = treedb_storage_t_Unlock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = treedb_storage_t_WriteLock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	TreedbInsertResult res2 = IntTree_Insert
		(& runtime->u_tree.m_tree, item);
	switch(res2) {
	case eTreedbInsertSuccess:
		return eCometCommandSuccess;
	case eTreedbInsertExists:
		return eCometCommandFailure;
	default:
	case eTreedbInsertNoMem:
		return eCometCommandOutOfResources;
	}
}
CometCommandStatus int_tree_remove
( meta_comet_ptr_t runtime
, const char * szitem
) {
	char * tailptr;
	int item = (int)strtol(szitem, & tailptr, 10);
	if(!tailptr || *tailptr) {
		  fprintf(stderr, "Error \"%s\" isn't a number.\n", szitem);
		  return eCometCommandBadCommand;
	}
	if(int_tree_setup(runtime))
		return eCometCommandHeapError;

	TreedbHeapStatus res = treedb_storage_t_Unlock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = treedb_storage_t_WriteLock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	BOOL ret = IntTree_Remove
		(& runtime->u_tree.m_tree, item);
	return ret == 1 ? eCometCommandSuccess : eCometCommandFailure;
}
CometCommandStatus int_tree_contains
( meta_comet_ptr_t runtime
, const char * szitem)
{
	char * tailptr;
	int item = (int)strtol(szitem, & tailptr, 10);
	if(!tailptr || *tailptr) {
		  fprintf(stderr, "Error \"%s\" isn't a number.\n", szitem);
		  return eCometCommandBadCommand;
	}
	if(int_tree_setup(runtime))
		return eCometCommandHeapError;
	META_TREEDB_printf2("Node top: %zu, item offset %zu.\n"
		, (size_t)treedb_tree_t_Top(& runtime->u_tree.m_tree)
		, (size_t)runtime->u_tree.m_description.m_aligned_node_size
	);
	BOOL ret = IntTree_Contains(& runtime->u_tree.m_tree, item);
	return ret == 1 ? eCometCommandSuccess : eCometCommandFailure;
}
CometCommandStatus int_tree_stream
( meta_comet_ptr_t runtime
) {
	if(int_tree_setup(runtime))
		return eCometCommandHeapError;
	TreedbHeapStatus res = treedb_storage_t_Unlock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = treedb_storage_t_WriteLock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	int32_t n;
	CometCommandStatus ret = eCometCommandSuccess;
	int c;
	for(;;) {
		c = scanf("%d", &n);
		if(c != 1)
			break;
		ret = (IntTree_Insert(& runtime->u_tree.m_tree, n)
				 == eTreedbInsertNoMem)
			? eCometCommandOutOfResources : eCometCommandSuccess;
		if(ret == eCometCommandOutOfResources)
			break;
	}
	return ret;
}
CometCommandStatus int_tree_type
( meta_comet_ptr_t runtime
, const char * filename
) {
	if(int_tree_setup(runtime))
		return eCometCommandHeapError;
	/*int bad = treedb_allocator_t_Check(& runtime->m_allocator);*/
	int bad = 0;
	char u[37];
	uuid_unparse_lower(runtime->m_storage.m_pTreedb->m_uSchema, u);
	if(bad) {
		printf(
			"%s is a %u bit int set treedb [%s] [check failed - %d].\n"
			, filename, runtime->m_storage.m_pTreedb->m_uBytesPerOffset*8, u, bad);
	}else{
		size_t count = treedb_tree_t_Count(& runtime->u_tree.m_tree);
		printf(
			"%s is a %u bit int set treedb [%s] with %zu entries.\n"
			, filename, runtime->m_storage.m_pTreedb->m_uBytesPerOffset*8, u
			, count);
	}
	fflush(stdout);
	treedb_allocator_t_DumpSchema(& runtime->m_allocator, STDOUT_FILENO);
	return eCometCommandSuccess;
}

CometDetails int_tree_details =
{ UUID_treedb_int_tree_schema
, & int_tree_init
, 3
};

CometCommandStatus int_tree_process
( meta_comet_ptr_t runtime
, int argc, char *argv[]
) {
	switch(argc) {
	case 6:
		/* prg set string <op> <filename> arg */
		if(!strcmp(argv[3], "insert"))
			return int_tree_insert(runtime, argv[5]);
		if(!strcmp(argv[3], "remove"))
			return int_tree_remove(runtime, argv[5]);
		if(!strcmp(argv[3], "contains"))
			return int_tree_contains(runtime, argv[5]);
		break;
	}
	return eCometCommandUnrecognisedCommand;
}
