/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <stddef.h>
#include "meta-comet.h"
#include "string-list.h"
#include <treedb/comet-guids.h>

CometCommandStatus string_list_setup
( meta_comet_ptr_t runtime
) {
	treedb_const_list_ptr_t plist_template
		= treedb_get_list_template_for
			(runtime->m_storage.m_pTreedb->m_uBytesPerOffset
			, eTreedbContainerHasCount);
	if(!plist_template) {
		fprintf(stderr, "No matching list template found.\n");
		return 1;
	}
	treedb_list_backend_t_Setup
		( & runtime->u_list.m_backend
		, & StringList_backend_vtbl
		);
	treedb_list_impl_t_Setup
		( & runtime->u_list.m_list_impl
		, & runtime->m_allocator
		, & runtime->u_list.m_backend
		, & runtime->u_list.m_description
		, plist_template->m_pImpl
		);
	/* We want to be compatible with treedb by using every last byte. */
	runtime->u_list.m_description.m_aligned_node_size
		= plist_template->m_pImpl->m_pDescription->m_node.m_uSize;

	treedb_list_t_Setup
		( & runtime->u_list.m_list
		, & runtime->u_list.m_list_impl
		, runtime->m_storage.m_pTreedb->m_ulPayloadOffset
		);
	return 0;
}
CometCommandStatus string_list_init
( meta_comet_ptr_t runtime
) {
	if(string_list_setup(runtime))
		return eCometCommandHeapError;
	treedb_list_t_Init(& runtime->u_list.m_list);
	return eCometCommandSuccess;
}

CometCommandStatus string_list_dump
( meta_comet_ptr_t runtime
) {
	if(string_list_setup(runtime))
		return eCometCommandHeapError;
	/*L2StringList_Dump(& runtime->u_list.m_list);*/
	treedb_list_ptr_t plist = & runtime->u_list.m_list;
	PVOID node = treedb_list_t_PointerHead(plist);
	for(; node; node = treedb_list_t_PointerNodeNextPtr(plist, node)) {
		const char * p = StringList_Data(plist, node);
		printf("%s\n", p);
	}
	return eCometCommandSuccess;
}
CometCommandStatus string_list_count
( meta_comet_ptr_t runtime
) {
	if(string_list_setup(runtime))
		return eCometCommandHeapError;
	size_t count = treedb_list_t_Count(& runtime->u_list.m_list);
	printf("%zu\n", count);
	return eCometCommandSuccess;
}
CometCommandStatus string_list_push_back
( meta_comet_ptr_t runtime
, const char * szitem
) {
	if(string_list_setup(runtime))
		return eCometCommandHeapError;

	TreedbHeapStatus res = treedb_storage_t_Unlock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = treedb_storage_t_WriteLock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	HTREEDB_LISTNODE res2 = StringList_PushBack
		(& runtime->u_list.m_list, szitem);
	if(!res2)
		return eCometCommandOutOfResources;
	return eCometCommandSuccess;
}
CometCommandStatus string_list_push_front
( meta_comet_ptr_t runtime
, const char * szitem
) {
	if(string_list_setup(runtime))
		return eCometCommandHeapError;

	TreedbHeapStatus res = treedb_storage_t_Unlock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = treedb_storage_t_WriteLock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	HTREEDB_LISTNODE res2 = StringList_PushFront
		(& runtime->u_list.m_list, szitem);
	if(!res2)
		return eCometCommandOutOfResources;
	return eCometCommandSuccess;
}
CometCommandStatus string_list_pop_back
( meta_comet_ptr_t runtime
) {
	if(string_list_setup(runtime))
		return eCometCommandHeapError;

	TreedbHeapStatus res = treedb_storage_t_Unlock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = treedb_storage_t_WriteLock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	BOOL ret = StringList_PopBack(& runtime->u_list.m_list);
	return ret == 0 ? eCometCommandFailure : eCometCommandSuccess;
}
CometCommandStatus string_list_pop_front
( meta_comet_ptr_t runtime
) {
	if(string_list_setup(runtime))
		return eCometCommandHeapError;

	TreedbHeapStatus res = treedb_storage_t_Unlock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = treedb_storage_t_WriteLock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	BOOL ret = StringList_PopFront(& runtime->u_list.m_list);
	return ret == 0 ? eCometCommandFailure : eCometCommandSuccess;
}
CometCommandStatus string_list_stream
( meta_comet_ptr_t runtime
) {
	if(string_list_setup(runtime))
		return eCometCommandHeapError;
	TreedbHeapStatus res = treedb_storage_t_Unlock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = treedb_storage_t_WriteLock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	char * p;
	CometCommandStatus ret = eCometCommandSuccess;
	int c;
	for(;;) {
		c = scanf("%as", &p);
		if(c != 1)
			break;
		ret = (StringList_PushBack(& runtime->u_list.m_list, p)
				 == 0)
			? eCometCommandOutOfResources : eCometCommandSuccess;
		free(p);
		if(ret == eCometCommandOutOfResources)
			break;
	}
	return ret;
}
CometCommandStatus string_list_type
( meta_comet_ptr_t runtime
, const char * filename
) {
	if(string_list_setup(runtime))
		return eCometCommandHeapError;
	/*int bad = treedb_allocator_t_Check(& runtime->m_allocator);*/
	int bad = 0;
	char u[37];
	uuid_unparse_lower(runtime->m_storage.m_pTreedb->m_uSchema, u);
	if(bad) {
		printf(
			"%s is a %u bit string list treedb [%s] [check failed - %d].\n"
			, filename, runtime->m_storage.m_pTreedb->m_uBytesPerOffset*8, u, bad);
	}else{
		size_t count = treedb_list_t_Count(& runtime->u_list.m_list);
		printf(
			"%s is a %u bit string list treedb [%s] with %zu entries.\n"
			, filename, runtime->m_storage.m_pTreedb->m_uBytesPerOffset*8, u
			, count);
	}
	fflush(stdout);
	treedb_allocator_t_DumpSchema(& runtime->m_allocator, STDOUT_FILENO);
	return eCometCommandSuccess;
}

CometDetails string_list_details =
{ UUID_treedb_string_list_schema
, & string_list_init
, 3
};

CometCommandStatus string_list_process
( meta_comet_ptr_t runtime
, int argc, char *argv[]
) {
	switch(argc) {
	case 5:
		/* prg set string <op> <filename> */
		if(!strcmp(argv[3], "pop_back"))
			return string_list_pop_back(runtime);
		if(!strcmp(argv[3], "pop_front"))
			return string_list_pop_front(runtime);
		break;
	case 6:
		/* prg set string <op> <filename> arg */
		if(!strcmp(argv[3], "push_back"))
			return string_list_push_back(runtime, argv[5]);
		if(!strcmp(argv[3], "push_front"))
			return string_list_push_front(runtime, argv[5]);
		break;
	}
	return eCometCommandUnrecognisedCommand;
}
