/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <stddef.h>
#include "meta-comet.h"
#include "string-tree.h"

CometCommandStatus string_tree_setup
( meta_comet_ptr_t runtime
) {
	treedb_const_tree_ptr_t ptree_template
		= treedb_get_avl_tree_template_for
			(runtime->m_storage.m_pTreedb->m_uBytesPerOffset
			/* Map meta-treedb enum values to treedb ones : multiply by four. */
			, runtime->m_storage.m_pTreedb->m_uFeatures * 4);
	if(!ptree_template) {
		fprintf(stderr, "No matching tree template found.\n");
		return 1;
	}
	treedb_tree_backend_t_Setup
		( & runtime->u_tree.m_backend
		, & StringTree_backend_vtbl
		);
	treedb_tree_impl_t_Setup2
		( & runtime->u_tree.m_tree_impl
		, & runtime->m_allocator
		, & runtime->u_tree.m_backend
		, & runtime->u_tree.m_description
		, ptree_template->m_pImpl
		);
	/* We want to be compatible with treedb by using every last byte. */
	META_TREEDB_printf2("Changing aligned_node_size from %zu to %zu.\n"
		, (size_t)runtime->u_tree.m_description.m_aligned_node_size
		, (size_t)ptree_template->m_pImpl->m_pDescription->m_node.m_uSize
	);
	runtime->u_tree.m_description.m_aligned_node_size
		= ptree_template->m_pImpl->m_pDescription->m_node.m_uSize;

	treedb_tree_t_Setup
		( & runtime->u_tree.m_tree
		, & runtime->u_tree.m_tree_impl
		, runtime->m_storage.m_pTreedb->m_ulPayloadOffset
		);
	return 0;
}
CometCommandStatus string_tree_init
( meta_comet_ptr_t runtime
) {
	if(string_tree_setup(runtime))
		return eCometCommandHeapError;
	treedb_tree_t_Init(& runtime->u_tree.m_tree);
	return eCometCommandSuccess;
}

CometCommandStatus string_tree_dump
( meta_comet_ptr_t runtime
) {
	if(string_tree_setup(runtime))
		return eCometCommandHeapError;
	/*AvlStringTree_Dump(& runtime->u_tree.m_tree);*/
	treedb_tree_ptr_t ptree = & runtime->u_tree.m_tree;
	PVOID node = treedb_tree_t_PointerHead(ptree);
	for(; node; node = treedb_tree_t_PointerNodeNextPtr(ptree, node)) {
		const char * p = StringTree_Data(ptree, node);
		printf("%s\n", p);
	}
	return eCometCommandSuccess;
}
CometCommandStatus string_tree_count
( meta_comet_ptr_t runtime
) {
	if(string_tree_setup(runtime))
		return eCometCommandHeapError;
	size_t count = treedb_tree_t_Count(& runtime->u_tree.m_tree);
	printf("%zu\n", count);
	return eCometCommandSuccess;
}
CometCommandStatus string_tree_insert
( meta_comet_ptr_t runtime
, const char * szitem
) {
	if(string_tree_setup(runtime))
		return eCometCommandHeapError;

	TreedbHeapStatus res = treedb_storage_t_Unlock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = treedb_storage_t_WriteLock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	TreedbInsertResult res2 = StringTree_Insert
		(& runtime->u_tree.m_tree, szitem);
	switch(res2) {
	case eTreedbInsertSuccess:
		return eCometCommandSuccess;
	case eTreedbInsertExists:
		return eCometCommandFailure;
	default:
	case eTreedbInsertNoMem:
		return eCometCommandOutOfResources;
	}
}
CometCommandStatus string_tree_remove
( meta_comet_ptr_t runtime
, const char * szitem
) {
	if(string_tree_setup(runtime))
		return eCometCommandHeapError;

	TreedbHeapStatus res = treedb_storage_t_Unlock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = treedb_storage_t_WriteLock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	BOOL ret = StringTree_Remove
		(& runtime->u_tree.m_tree, szitem);
	return ret != 0 ? eCometCommandSuccess : eCometCommandFailure;
}
CometCommandStatus string_tree_contains
( meta_comet_ptr_t runtime
, const char * szitem)
{
	if(string_tree_setup(runtime))
		return eCometCommandHeapError;

	BOOL ret = StringTree_Contains
		(& runtime->u_tree.m_tree, szitem);
	return ret == 1 ? eCometCommandSuccess : eCometCommandFailure;
}
CometCommandStatus string_tree_stream
( meta_comet_ptr_t runtime
) {
	if(string_tree_setup(runtime))
		return eCometCommandHeapError;
	TreedbHeapStatus res = treedb_storage_t_Unlock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = treedb_storage_t_WriteLock(& runtime->m_storage);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock : %s.\n"
			, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	char * p;
	CometCommandStatus ret = eCometCommandSuccess;
	int c;
	for(;;) {
		c = scanf("%as", &p);
		if(c != 1)
			break;
		ret = (StringTree_Insert(& runtime->u_tree.m_tree, p)
				 == eTreedbInsertNoMem)
			? eCometCommandOutOfResources : eCometCommandSuccess;
		free(p);
		if(ret == eCometCommandOutOfResources)
			break;
	}
	return ret;
}
CometCommandStatus string_tree_type
( meta_comet_ptr_t runtime
, const char * filename
) {
	if(string_tree_setup(runtime))
		return eCometCommandHeapError;
	/*int bad = treedb_allocator_t_Check(& runtime->m_allocator);*/
	int bad = 0;
	char u[37];
	uuid_unparse_lower(runtime->m_storage.m_pTreedb->m_uSchema, u);
	if(bad) {
		printf(
			"%s is a %u bit string set treedb [%s] [check failed - %d].\n"
			, filename, runtime->m_storage.m_pTreedb->m_uBytesPerOffset*8, u, bad);
	}else{
		size_t count = treedb_tree_t_Count(& runtime->u_tree.m_tree);
		printf(
			"%s is a %u bit string set treedb [%s] with %zu entries.\n"
			, filename, runtime->m_storage.m_pTreedb->m_uBytesPerOffset*8, u
			, count);
	}
	fflush(stdout);
	treedb_allocator_t_DumpSchema(& runtime->m_allocator, STDOUT_FILENO);
	return eCometCommandSuccess;
}

CometDetails string_tree_details =
{ UUID_treedb_string_tree_schema
, & string_tree_init
, 3
};

CometCommandStatus string_tree_process
( meta_comet_ptr_t runtime
, int argc, char *argv[]
) {
	switch(argc) {
	case 6:
		/* prg set string <op> <filename> arg */
		if(!strcmp(argv[3], "insert"))
			return string_tree_insert(runtime, argv[5]);
		if(!strcmp(argv[3], "remove"))
			return string_tree_remove(runtime, argv[5]);
		if(!strcmp(argv[3], "contains"))
			return string_tree_contains(runtime, argv[5]);
		break;
	}
	return eCometCommandUnrecognisedCommand;
}
