/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/** \example int-tree.c int tree treedb example
In this example the tree "payload" is a 32 bit signed integer.

int-tree.c and int-tree.h provide a backend for a treedb_t so that it can manage
int32_t tree payloads.

The added complication this example addresses is the alignment requirement of
the payload, in addition to the alignment requirements of the tree node.
*/

#include "int-tree.h"
#include <stdio.h>
#include <unistd.h>
#include <v3c/meta-avl.h> /* We need to start from somewhere. */

IntTreeData g_data;

int setup_IntTreeData
( uint16_t abytes
, uint16_t aflags
, uint16_t xbytes
, uint16_t xflags
) {
	{
		treedb_storage_on_memory_t stg
			= TREEDB_STORAGE_ON_MEMORY_INIT(g_buf(), 4096);
		g_data.storage = stg;
	}
	/* The allocator. */
	g_data.allocator.m_pImpl = & g_data.allocator_impl;
	if(!treedb_init_allocator_on_storage_from_template
			(& g_data.allocator, & g_data.allocator_impl
			, abytes, aflags, sizeof(TreedbHeader)
			, (treedb_storage_ptr_t)(& g_data.storage))) {
		printf("treedb_init_allocator_on_storage_from_template failed.\n");
		return 1;
	}
	g_data.ptree_template = treedb_get_avl_tree_template_for(xbytes, xflags);
	if(!g_data.ptree_template) {
		printf("No matching tree template found.\n");
		return 2;
	}
	if(!treedb_allocator_t_Init(& g_data.allocator
			, g_data.ptree_template->m_pImpl->m_pDescription->m_root.m_uSize
			, 4096
			, & g_data.payload)) {
		printf("Allocator Init() failed.\n");
		return 3;
	}
	printf("Header: %u, Allocator: %zu, Node: %u, Item %u, Node total %u\n"
		, (unsigned int)sizeof(TreedbHeader)
		, (size_t)(g_data.payload - sizeof(TreedbHeader))
		, g_data.allocator.m_free.m_pImpl->m_pDescription->m_node.m_uSize
		, g_data.allocator.m_pImpl->m_pDescription->m_aligned_item_size
		, g_data.allocator.m_free.m_pImpl->m_pDescription->m_aligned_node_size
			+ g_data.allocator.m_pImpl->m_pDescription->m_aligned_item_size
	);

	if(!treedb_allocator_t_LargestFree(&g_data.allocator, &g_data.free_offset
			, &g_data.free_size)) {
		printf("treedb_allocator_LargestFree failed.\n");
		return 4;
	}

	/* Set up the int tree backend. */
	treedb_tree_backend_t_Setup(& g_data.backend, & IntTree_backend_vtbl);

	/* Set up the int tree implementation. */
	treedb_tree_impl_t_Setup2
		( & g_data.int_tree_impl
		, & g_data.allocator
		, & g_data.backend
		, & g_data.description
		, g_data.ptree_template->m_pImpl
		);
	/* We need to align the int32_t on a 4-byte boundary, like the compiler
		does, to be compatible with treedb, and as the backend expects it.
	*/
	uint16_t nsz
		= V3C_ROUND_UP_T2(uint16_t
			, g_data.description.m_node.m_uSize
			, sizeof(int32_t)
			);
	META_TREEDB_printf2("Changing aligned_node_size from %zu to %u because of "
		"node size %u.\n"
		, (size_t)g_data.description.m_aligned_node_size
		, nsz, g_data.description.m_node.m_uSize);
	g_data.description.m_aligned_node_size = nsz;

	/* Set up the int tree. */
	if(xflags & eTreedbContainerIsPointer)
		g_data.payload
			= (uintptr_t)treedb_allocator_t_OffsetToPointer(& g_data.allocator
				, g_data.payload);
	treedb_tree_t_Setup
		( & g_data.int_tree
		, & g_data.int_tree_impl
		, g_data.payload
		);
	treedb_tree_t_Init(& g_data.int_tree);
	return 0;
}

static HTREEDB_TREEKEY STDCALL IntTree_UnpackKey
( treedb_tree_backend_ptr_t This
, treedb_tree_runtime_ptr_t runtime
, PCVOID pv
) {
	(void)This;
	return IntTree_DataPointer(runtime, pv);
}
static int STDCALL IntTree_Compare
( treedb_tree_backend_ptr_t This
, treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREEKEY av
, HTREEDB_TREEKEY bv
) {
	(void)This;
	(void)runtime;
	int32_t a = *(int32_t *)av;
	int32_t b = *(int32_t *)bv;
	int res = (a < b) ? -1 : (a > b) ? 1 : 0;
#if META_TREEDB_DEBUG
	int ain = treedb_allocator_t_InHeapPtr(runtime->m_pImpl->m_pAllocator, av);
	int bin = treedb_allocator_t_InHeapPtr(runtime->m_pImpl->m_pAllocator, bv);
	fprintf(stderr, "IntTree_Compare [");
	if(ain)
		fprintf(stderr, "%zu:%d", treedb_tree_t_PointerToOffset(runtime, av), a);
	else
		fprintf(stderr, "<external>:%d", a);
	fprintf(stderr, "] [");
	if(bin)
		fprintf(stderr, "%zu:%d", treedb_tree_t_PointerToOffset(runtime, bv), b);
	else
		fprintf(stderr, "<external>:%d", a);
	fprintf(stderr, "] : %d.\n", res);
#endif
	return res;
}
static PVOID STDCALL IntTree_AllocNode
( treedb_tree_backend_ptr_t This
, treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREEKEY pv
) {
	(void)This;
	uintptr_t size = runtime->m_pImpl->m_pDescription->m_aligned_node_size;
	char * ret = (char *)treedb_allocator_t_PointerAlloc
				( runtime->m_pImpl->m_pAllocator
				, size + sizeof(int32_t));
	if(!ret)
		return 0;
	size_t misalign = treedb_tree_t_PointerPack(runtime, ret + size)
		% sizeof(int32_t);
	META_TREEDB_printf2("IntTree_AllocNode : Alloc'd %zu, storing %d to %zu "
		", misalign %zu.\n"
		, treedb_tree_t_PointerToOffset(runtime, ret)
		, *(int32_t *)pv
		, treedb_tree_t_PointerToOffset(runtime, ret + size)
		, misalign
	);
	if(misalign) {
		treedb_allocator_t_FreePtr(runtime->m_pImpl->m_pAllocator, ret);
		return 0;
	}
	*(int32_t*)(ret + size) = *(int32_t *)pv;
	return ret;
}
static void STDCALL IntTree_FreeNode
( treedb_tree_backend_ptr_t This
, treedb_tree_runtime_ptr_t runtime
, PVOID pv
) {
	(void)This;
	treedb_allocator_t_FreePtr(runtime->m_pImpl->m_pAllocator, pv);
}
#if 0
#define ve(x) v->x = & IntTree_##x
static void test(treedb_tree_backend_vtbl_t * v)
{
	ve(UnpackKey);
	ve(Compare);
	ve(AllocNode);
	ve(FreeNode);
}
#endif
treedb_tree_backend_vtbl_t IntTree_backend_vtbl = {
	treedb_tree_backend_vtbl_t_entries(IntTree_)
};

typedef struct IntTree_tree_data_renderer_t IntTree_tree_data_renderer_t
, * IntTree_tree_data_renderer_ptr_t;
typedef struct IntTree_tree_data_renderer_vtbl_t
	IntTree_tree_data_renderer_vtbl_t;
typedef const IntTree_tree_data_renderer_vtbl_t
	* IntTree_tree_data_renderer_const_vtbl_ptr_t;
struct IntTree_tree_data_renderer_vtbl_t
{
	treedb_tree_data_renderer_t_methods(,,IntTree_tree_data_renderer_t);
};
struct IntTree_tree_data_renderer_t
{
	treedb_tree_data_renderer_t_members(
		IntTree_tree_data_renderer_const_vtbl_ptr_t);
	uint16_t m_aligned_node_size;
};

static PSTR STDCALL IntTree_data_Dump
( IntTree_tree_data_renderer_ptr_t This
, treedb_tree_runtime_ptr_t runtime
, PCVOID pv
) {
	(void)runtime;
	const char * p = ((const char *)pv) + This->m_aligned_node_size;
	fprintf(stderr, "IntTree_data_Dump : Reading \"%d\" from %zu.\n"
		, *(int32_t *)p, treedb_tree_t_PointerToOffset(runtime, p));
	fflush(stdout);
	char * ret = 0;
	if(0 > asprintf(& ret, "%d", *(int32_t *)p))
		return 0;
	return ret;
}
static IntTree_tree_data_renderer_vtbl_t IntTree_tree_data_renderer_vtbl = {
	treedb_tree_data_renderer_vtbl_t_entries(IntTree_data_)
};

void STDCALL IntTree_Dump
( treedb_tree_ptr_t This
) {
	IntTree_tree_data_renderer_t IntTree_tree_data_renderer =
	{ & IntTree_tree_data_renderer_vtbl
	, This->m_pImpl->m_pDescription->m_aligned_node_size
	};
	treedb_fd_tree_renderer_t renderer = INIT_TREEDB_FD_TREE_RENDERER(STDOUT_FILENO);
	treedb_allocator_t_Dump(This->m_pImpl->m_pAllocator
		, (treedb_tree_renderer_ptr_t)& renderer
		, (treedb_tree_data_renderer_ptr_t)& IntTree_tree_data_renderer);
}
