/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <iostream>
#include <cstdio> /* asprintf() */
#include <v3c/meta-avl.h> /* We need to start from somewhere. */
#include <v3c/meta-treedb.hpp>

#if HAVE_64_BITS
uint16_t bpo_end = 16;
#else
uint16_t bpo_end = 8;
#endif

using namespace std;
using namespace v3c;
using namespace treedb;
using namespace meta;

char buf[4096 + 8];
char * g_buf() { return V3C_ROUND_UP_PTR(buf,8); }

const char * flags_desc(unsigned int flags)
{
	switch(flags) {
	case eTreedbContainerHasCount:
		return "count";
	case eTreedbContainerHasDepth:
		return "depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth:
		return "count | depth";
	case eTreedbContainerHasCount | eTreedbContainerIsPointer:
		return "pointer | count";
	case eTreedbContainerHasDepth | eTreedbContainerIsPointer:
		return "pointer | depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth | eTreedbContainerIsPointer:
		return "pointer | count | depth";
	default:
		return "<unknown flags>";
	}
}

struct IntTreeBackendTypes {
	typedef int32_t KeyType;
	typedef int32_t DataType;
	typedef int32_t InputType;
};
static bool g_bad_align = false;
struct IntTreeBackend
: public FixedSimpleTreeBackendT
			< IntTreeBackend, TreeRuntime, IntTreeBackendTypes, sizeof(int32_t) >
{
	inline static const int32_t & unpack_data(PCVOID pv)
	{ return *(const int32_t*)pv; }
	inline static int32_t & unpack_data(PVOID pv)
	{ return *(int32_t*)pv; }
	inline static HTREEDB_TREEKEY pack_input(const int32_t & i)
	{ return (HTREEDB_TREEKEY)(&i); }

	inline int compare(int32_t a, int32_t b)
	{ return (a < b) ? -1 : (a > b) ? 1 : 0; }
	inline static char * render(int32_t i)
	{
		if(i > 10)
			g_bad_align = true;
		char * ret;
		if( 0 > asprintf(& ret, "%d", i))
			return 0;
		return ret;
	}
};
/*! A treedb_tree_t delegate.
*/
typedef TreeImplT< IntTreeBackend > IntTreeImpl;
typedef TreeWithImplT< IntTreeBackend > IntTreeWithImpl;

int test(uint16_t abytes, uint16_t aflags, uint16_t xbytes, uint16_t xflags)
{
	cout << "############################################################"
		"####################\n";
	if(aflags & eTreedbContainerIsPointer)
		cout << "# "
#ifdef __cplusplus
		"C++ "
#endif
		"Int tree : allocator [" << flags_desc(aflags) << "], ";
	else
		cout << "# "
#ifdef __cplusplus
		"C++ "
#endif
		"Int tree : allocator [" << abytes*8 << " : " << flags_desc(aflags)
			<< "], ";
	if(xflags & eTreedbContainerIsPointer)
		cout << "tree[" << flags_desc(xflags) << "].\n";
	else
		cout << "tree[" << xbytes*8 << " : " << flags_desc(xflags) << "].\n";
	cout << "############################################################"
		"####################" << endl;

	/* The storage. */
	StorageOnMemory storage(g_buf(), 4096);

	/* The allocator. */
	AllocatorOnStorageImpl allocator_impl;
	AllocatorOnStorage allocator;
	allocator.m_inst.m_pImpl = & allocator_impl.m_inst;

	/* Even though only the size member is needed from the header, I'm including
	it to compare the numbers with meta-comet. */
	if(!allocator.init_runtime(& allocator_impl.c_type(), abytes, aflags
			, sizeof(TreedbHeader), & storage.base_instance())) {
		cerr << "No matching allocator template found." << endl;
		return 1;
	}
	treedb_const_tree_ptr_t ptree_template
		= treedb_get_avl_tree_template_for(xbytes, xflags);
	if(!ptree_template) {
		cerr << "No matching tree template found." << endl;
		return 1;
	}
	PackedPayload< IntTreeImpl > payload;
	uint64_t payload_64;

	if(!allocator.Init
			( ptree_template->m_pImpl->m_pDescription->m_root.m_uSize
			, 4096
			, & payload_64))
		return 1;
	if(payload_64 > (uint64_t)((size_t)-1))
		return 1;
	payload.m_value = (size_t)payload_64;

	cout << "Header: " << sizeof(TreedbHeader)
		<< ", Allocator: " << payload.m_value - sizeof(TreedbHeader)
		<< ", Node: " <<
				allocator.free_tree().m_pImpl->m_pDescription->m_node.m_uSize
		<< ", offset: " <<
				allocator.free_tree().m_pImpl->m_pDescription->m_pointer.m_uSize
		<< ", Item " << allocator.description().m_aligned_item_size
		<< ", Node total " <<
				allocator.free_tree().m_pImpl->m_pDescription->m_node.m_uSize
				+ allocator.description().m_aligned_item_size
		<< "\n";

	size_t free_offset, free_size;
	if(!allocator.LargestFree(&free_offset, &free_size)) {
		cerr << "treedb_allocator_LargestFree failed." << endl;
		return 1;
	}

	if(xflags & eTreedbContainerIsPointer)
		payload.m_value = (uintptr_t)allocator.OffsetToPointer(payload.m_value);
	IntTreeWithImpl tree(allocator, ptree_template, payload);
	tree.Init();
	unsigned int n;

	tree.dump();

	treedb_allocator_t_check_result
		( & allocator.base_c_type()
		, allocator.Check()
	);
	if(allocator.error()) {
		cerr << "Allocator check failed : " << allocator.error() << "." << endl;
		return 6;
	}
	for(n = 0; n < 10; ++n) {
		cout << "[" << n << "] inserting " << n << "." << endl;
		TreedbInsertResult res = tree.insert(n);
		if(res != eTreedbInsertSuccess) {
			cerr << "[" << n << "] insertion failed: "
				<< TreedbInsertResultDesc(res) << "." << endl;
			return 1;
		}
#if META_TREEDB_DEBUG
		if(!allocator.error())
			allocator.Check();
#endif
		if(allocator.error()) {
			cerr << "Allocator check failed : " << allocator.error() << "."
				<< endl;
			return 1;
		}
		tree.dump();
		if(g_bad_align) {
			cerr << "Data alignment check failed." << endl;
			return 1;
		}
	}
	--n;
	TreedbInsertResult res = tree.insert(n);
	if(res != eTreedbInsertExists) {
		tree.dump();
		if(g_bad_align) {
			cerr << "Data alignment check failed." << endl;
			return 1;
		}
		cerr << "[" << n << "] insertion should have failed: "
			<< TreedbInsertResultDesc(res) << "." << endl;
		return 1;
	}
	uintptr_t count = tree.Count();
	cout << "Inserted " << count << " items.\n";

	PVOID node = tree.PointerHead();

	for(n = 0; node; ++n, node = tree.PointerNodeNextPtr(node)) {
		int d = tree.data(node);
		cout << "[" << n << "] " << d << "\n";
		if(d != (int)n) {
			cerr << "[" << d << "] != " << n << "." << endl;
			return 1;
		}
	}

	tree.Clear();
	treedb_allocator_t_check_result
		( & allocator.base_c_type()
		, allocator.Check()
	);
	if(allocator.error()) {
		cerr << "Allocator check failed : " << allocator.error() << ".\n";
		return 1;
	}

	size_t new_free_offset, new_free_size;
	if(!allocator.LargestFree(&new_free_offset, &new_free_size)) {
		cerr << "Error : treedb_allocator_LargestFree(2) failed.\n";
		return 1;
	}
	if(free_size != new_free_size) {
		cerr << "Error : free size " << new_free_size
			<< " != original free size " << free_size << ".\n";
		return 1;
	}
	if(free_offset != new_free_offset) {
		cerr << "Error : free offset " << new_free_offset
			<< " != original free offset " << free_offset << ".\n";
		return 1;
	}
	if(count != 10) {
		cerr << "Error : Expected a count of 10, got " << count << " instead.\n";
		return 1;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	unsigned int flags[] =
	{ eTreedbContainerHasCount
	, eTreedbContainerHasDepth
	, eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int pflags[] =
	{ eTreedbContainerIsPointer | eTreedbContainerHasCount
	, eTreedbContainerIsPointer | eTreedbContainerHasDepth
	, eTreedbContainerIsPointer | eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int abytes, aflags_index;
	unsigned int xbytes, xflags_index;
	/* Work through the allocator byte range */
	for(abytes = 2; abytes < bpo_end; abytes <<= 1) {
		/* Work through the allocator flags */
		for(aflags_index = 0
				; aflags_index < sizeof(flags)/sizeof(flags[0])
				; ++aflags_index) {
			/* Work through the tree byte range */
			for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
				/* Work through the tree flags */
				for(xflags_index = 0
						; xflags_index < sizeof(flags)/sizeof(flags[0])
						; ++xflags_index) {
					int ret = test(abytes, flags[aflags_index], xbytes
							, flags[xflags_index]);
					if(ret)
						return ret;
				}
			}
			/* Work through the pointer tree flags */
			for(xflags_index = 0;
					xflags_index < sizeof(pflags)/sizeof(pflags[0])
					; ++xflags_index) {
				int ret = test(abytes, flags[aflags_index], 0
						, pflags[xflags_index]);
				if(ret)
					return ret;
			}
		}
	}
	/* Work through the pointer allocator flags */
	for(aflags_index = 0;
			aflags_index < sizeof(pflags)/sizeof(pflags[0])
			; ++aflags_index) {
		/* Work through the tree byte range */
		for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
			/* Work through the tree flags */
			for(xflags_index = 0
					; xflags_index < sizeof(flags)/sizeof(flags[0])
					; ++xflags_index) {
				int ret = test(0, pflags[aflags_index], xbytes
							, flags[xflags_index]);
				if(ret)
					return ret;
			}
		}
		/* Work through the pointer tree flags */
		for(xflags_index = 0;
				xflags_index < sizeof(pflags)/sizeof(pflags[0])
				; ++xflags_index) {
			int ret = test(0, pflags[aflags_index], 0, pflags[xflags_index]);
			if(ret)
				return ret;
		}
	}
	cout << "Success!\n";
	return 0;
}
