/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <iostream>
#include <cstdio> // asprintf
#include <v3c/meta-avl.h> /* We need to start from somewhere. */
#include <v3c/meta-treedb.hpp>

#if HAVE_64_BITS
uint16_t bpo_end = 16;
#else
uint16_t bpo_end = 8;
#endif

using namespace std;
using namespace v3c;
using namespace treedb;
using namespace meta;

struct Data {
	int32_t m_data;
	char m_key[];
};
struct Input {
	const char * m_key;
	int32_t m_data;
};

Input sorted_items[] =
{ { "apple" , 0 }
, { "banana", 1 }
, { "lemon" , 2 }
, { "orange", 3 }
, { "pear"  , 4 }
, { "plum"  , 5 }
};
Input items[] =
{ { "apple" , 0 }
, { "orange", 3 }
, { "lemon" , 2 }
, { "pear"  , 4 }
, { "plum"  , 5 }
, { "banana", 1 }
};

char buf[4096 + 8];
char * g_buf() { return V3C_ROUND_UP_PTR(buf,8); }

const char * flags_desc(unsigned int flags)
{
	switch(flags) {
	case eTreedbContainerHasCount:
		return "count";
	case eTreedbContainerHasDepth:
		return "depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth:
		return "count | depth";
	case eTreedbContainerHasCount | eTreedbContainerIsPointer:
		return "pointer | count";
	case eTreedbContainerHasDepth | eTreedbContainerIsPointer:
		return "pointer | depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth | eTreedbContainerIsPointer:
		return "pointer | count | depth";
	default:
		return "<unknown flags>";
	}
}

struct StringTreeBackendTypes {
	typedef const char * KeyType;
	typedef Data DataType;
	typedef Input InputType;
};
struct StringTreeBackend
: public TreeBackendT
			< StringTreeBackend, TreeRuntime, StringTreeBackendTypes
, sizeof(char) >
{
	enum { eKeyOffset = sizeof(int32_t) };
	inline static const char * unpack_key(PCVOID pv)
	{ return ((const Data *)pv)->m_key; }
	inline static const Data & unpack_data(PCVOID pv)
	{ return *(const Data *)pv; }
	inline static Data & unpack_data(PVOID pv)
	{ return *(Data *)pv; }
	inline static const Input & unpack_input(PCVOID pv)
	{ return *(const Input *)pv; }

	inline static HTREEDB_TREEKEY pack_input(const Input & i)
	{ return (HTREEDB_TREEKEY)(&i); }
	inline static HTREEDB_TREEKEY pack_key(const char * k)
	{ return (HTREEDB_TREEKEY)((uintptr_t)k - eKeyOffset); }

	uintptr_t size(TreeRuntime *, const Input & in)
	{ return sizeof(Data) + strlen(in.m_key) + 1; }
	bool assign(TreeRuntime *, Data & to, const Input & from)
	{
		to.m_data = from.m_data;
		strcpy(to.m_key, from.m_key);
		return true;
	}
	inline int compare(const char * a, const char * b)
	{
		int ret = strcmp(a,b);
		cout << "StringTreeBackend::compare(" << a << ", " << b << ") == "
			<< ret << '.' << endl;
		return ret;
	}
	inline static char * render(const Data & i)
	{
		char * ret;
		if(asprintf(& ret, "[ %u, %s ]", i.m_data, i.m_key) < 0)
			return 0;
		return ret;
	}
};

/*! A treedb_tree_t delegate.
*/
typedef TreeImplT< StringTreeBackend > StringTreeImpl;
typedef TreeWithImplT< StringTreeBackend > StringTreeWithImpl;

int test(uint16_t abytes, uint16_t aflags, uint16_t xbytes, uint16_t xflags)
{
	cout << "############################################################"
		"####################\n";
	if(aflags & eTreedbContainerIsPointer)
		cout << "# "
#ifdef __cplusplus
		"C++ "
#endif
		"String tree : allocator [" << flags_desc(aflags) << "], ";
	else
		cout << "# "
#ifdef __cplusplus
		"C++ "
#endif
		"String tree : allocator [" << abytes*8 << " : " << flags_desc(aflags) << "], ";
	if(xflags & eTreedbContainerIsPointer)
		cout << "tree[" << flags_desc(xflags) << "].\n";
	else
		cout << "tree[" << xbytes*8 << " : " << flags_desc(xflags) << "].\n";
	cout << "############################################################"
		"####################" << endl;

	/* The storage. */
	StorageOnMemory storage(g_buf(), 4096);

	/* The allocator. */
	AllocatorOnStorageImpl allocator_impl;
	AllocatorOnStorage allocator;
	allocator.m_inst.m_pImpl = & allocator_impl.m_inst;

	/* Even though only the size member is needed from the header, I'm including
	it to compare the numbers with meta-comet. */
	if(!allocator.init_runtime(& allocator_impl.c_type(), abytes, aflags
			, sizeof(TreedbHeader), & storage.base_instance())) {
		cerr << "Error 1: No matching allocator template found." << endl;
		return 1;
	}
	treedb_const_tree_ptr_t ptree_template
		= treedb_get_avl_tree_template_for(xbytes, xflags);
	if(!ptree_template) {
		cerr << "Error 2: No matching tree template found." << endl;
		return 2;
	}
	PackedPayload< StringTreeImpl > payload;
	uint64_t payload_64;

	if(!allocator.Init
			( ptree_template->m_pImpl->m_pDescription->m_root.m_uSize
			, 4096
			, & payload_64)) {
		cerr << "Error 3: allocator.Init() failed." << endl;
		return 3;
	}
	if(payload_64 > (uint64_t)((size_t)-1))
		return 4;
	payload.m_value = (size_t)payload_64;

	cout << "Header: " << sizeof(TreedbHeader)
		<< ", Allocator: " << payload.m_value - sizeof(TreedbHeader)
		<< ", Node: " <<
				allocator.free_tree().m_pImpl->m_pDescription->m_node.m_uSize
		<< ", offset: " <<
				allocator.free_tree().m_pImpl->m_pDescription->m_pointer.m_uSize
		<< ", Item " << allocator.description().m_aligned_item_size
		<< ", Node total " <<
				allocator.free_tree().m_pImpl->m_pDescription->m_node.m_uSize
				+ allocator.description().m_aligned_item_size
		<< "\n";

	size_t free_offset, free_size;
	if(!allocator.LargestFree(&free_offset, &free_size)) {
		cerr << "Error 5: treedb_allocator_LargestFree failed." << endl;
		return 5;
	}

	if(xflags & eTreedbContainerIsPointer)
		payload.m_value = (uintptr_t)allocator.OffsetToPointer(payload.m_value);
	StringTreeWithImpl runtime(allocator, ptree_template, payload);
	runtime.Init();
	unsigned int n;

	runtime.dump();

	treedb_allocator_t_check_result
		( & allocator.base_c_type()
		, allocator.Check()
	);
	if(allocator.error()) {
		cerr << "Error 6: Allocator check failed : " << allocator.error()
			<< "." << endl;
		return 7;
	}
	for(n = 0; n < sizeof(items)/sizeof(items[0]); ++n) {
		cout << "[" << n << "] inserting " << items[n].m_key << "." << endl;
		TreedbInsertResult res = runtime.insert(items[n]);
		if(res != eTreedbInsertSuccess) {
			cerr << "Error 7: [" << n << "] insertion failed: "
				<< TreedbInsertResultDesc(res) << "." << endl;
			return 7;
		}
#if META_TREEDB_DEBUG
		if(!allocator.error())
			allocator.Check();
#endif
		if(allocator.error()) {
			cerr << "Error 8: Allocator check failed : " << allocator.error()
				<< "." << endl;
			return 8;
		}
		runtime.dump();
	}
	--n;
	TreedbInsertResult res = runtime.insert(items[n]);
	if(res != eTreedbInsertExists) {
		runtime.dump();
		cerr << "Error 9: [" << n << "] insertion should have failed: "
			<< TreedbInsertResultDesc(res) << "." << endl;
		return 9;
	}
	uintptr_t count = runtime.Count();
	cout << "Inserted " << count << " items.\n";


	PVOID node;
	const char * p;

	cout << "search_lt(\"apple\")" << endl;
	node = runtime.search_lt("apple");
	p = node ? runtime.data(node).m_key : 0;
	if(p) {
		cout << "Error 10: got \"" << p << "\" != <null>." << endl;
		return 10;
	}

	cout << "search_lt(\"plum0\")" << endl;
	node = runtime.search_lt("plum0");
	p = node ? runtime.data(node).m_key : "<null>";
	if(strcmp(p, "plum")) {
		cout << "Error 11: got \"" << p << "\" != \"plum\"." << endl;
		return 11;
	}

	cout << "search_lt(\"orange\")" << endl;
	node = runtime.search_lt("orange");
	p = node ? runtime.data(node).m_key : "<null>";
	if(strcmp(p, "lemon")) {
		cout << "Error 12: got \"" << p << "\" != \"lemon\"." << endl;
		return 12;
	}

	cout << "search_gt(\"plum\")" << endl;
	node = runtime.search_gt("plum");
	p = node ? runtime.data(node).m_key : 0;
	if(p) {
		cout << "Error 13: search_gt(\"plum\") : got \"" << p
			<< "\" != <null>." << endl;
		return 13;
	}

	cout << "search_gt(\"melon0\")" << endl;
	node = runtime.search_gt("melon0");
	p = node ? runtime.data(node).m_key : "<null>";
	if(strcmp(p, "orange")) {
		cout << "Error 14: got \"" << p << "\" != \"orange\"." << endl;
		return 14;
	}

	cout << "search_gt(\"grapefruit\")" << endl;
	node = runtime.search_gt("grapefruit");
	p = node ? runtime.data(node).m_key : "<null>";
	if(strcmp(p, "lemon")) {
		cout << "Error 15: got \"" << p << "\" != \"lemon\"." << endl;
		return 15;
	}

	node = runtime.PointerHead();

	for(n = 0; node; ++n, node = runtime.PointerNodeNextPtr(node)) {
		const char * p = runtime.data(node).m_key;
		int32_t i = runtime.data(node).m_data;
		cout << "[" << n << "] " << i << ", " << p
			<< endl;
		if(strcmp(p, sorted_items[n].m_key)) {
			cerr << "Error 16: [" << n << "] " << p << " != "
				<< sorted_items[n].m_key << "." << endl;
			return 16;
		}
		if(i != (int32_t)n) {
			cerr << "Error 16a: " << i << " != " << n << "." << endl;
			return 16;
		}
	}

	runtime.Clear();
	treedb_allocator_t_check_result
		( & allocator.base_c_type()
		, allocator.Check()
	);
	if(allocator.error()) {
		cerr << "Error 17: Allocator check failed : " << allocator.error()
			<< '.' << endl;
		return 17;
	}

	size_t new_free_offset, new_free_size;
	if(!allocator.LargestFree(&new_free_offset, &new_free_size)) {
		cerr << "Error 18: treedb_allocator_LargestFree(2) failed.\n";
		return 18;
	}
	if(free_size != new_free_size) {
		cerr << "Error 19: free size " << new_free_size
			<< " != original free size " << free_size << ".\n";
		return 19;
	}
	if(free_offset != new_free_offset) {
		cerr << "Error 20: free offset " << new_free_offset
			<< " != original free offset " << free_offset << ".\n";
		return 20;
	}
	if(count != sizeof(items)/sizeof(items[0])) {
		cerr << "Error 21: Expected a count of "
		<< sizeof(items)/sizeof(items[0])
		<< ", got " << count << " instead.\n";
		return 21;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	unsigned int flags[] =
	{ eTreedbContainerHasCount
	, eTreedbContainerHasDepth
	, eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int pflags[] =
	{ eTreedbContainerIsPointer | eTreedbContainerHasCount
	, eTreedbContainerIsPointer | eTreedbContainerHasDepth
	, eTreedbContainerIsPointer | eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int abytes, aflags_index;
	unsigned int xbytes, xflags_index;
	/* Work through the allocator byte range */
	for(abytes = 2; abytes < bpo_end; abytes <<= 1) {
		/* Work through the allocator flags */
		for(aflags_index = 0
				; aflags_index < sizeof(flags)/sizeof(flags[0])
				; ++aflags_index) {
			/* Work through the tree byte range */
			for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
				/* Work through the tree flags */
				for(xflags_index = 0
						; xflags_index < sizeof(flags)/sizeof(flags[0])
						; ++xflags_index) {
					int ret = test(abytes, flags[aflags_index], xbytes
							, flags[xflags_index]);
					if(ret)
						return ret;
				}
			}
			/* Work through the pointer tree flags */
			for(xflags_index = 0;
					xflags_index < sizeof(pflags)/sizeof(pflags[0])
					; ++xflags_index) {
				int ret = test(abytes, flags[aflags_index], 0
						, pflags[xflags_index]);
				if(ret)
					return ret;
			}
		}
	}
	/* Work through the pointer allocator flags */
	for(aflags_index = 0;
			aflags_index < sizeof(pflags)/sizeof(pflags[0])
			; ++aflags_index) {
		/* Work through the tree byte range */
		for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
			/* Work through the tree flags */
			for(xflags_index = 0
					; xflags_index < sizeof(flags)/sizeof(flags[0])
					; ++xflags_index) {
				int ret = test(0, pflags[aflags_index], xbytes
							, flags[xflags_index]);
				if(ret)
					return ret;
			}
		}
		/* Work through the pointer tree flags */
		for(xflags_index = 0;
				xflags_index < sizeof(pflags)/sizeof(pflags[0])
				; ++xflags_index) {
			int ret = test(0, pflags[aflags_index], 0, pflags[xflags_index]);
			if(ret)
				return ret;
		}
	}
	cout << "Success!\n";
	return 0;
}
