/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <iostream>
#include <cstdio> // asprintf
#include <v3c/meta-avl.h> /* We need to start from somewhere. */
#include <v3c/meta-treedb.hpp>

#if HAVE_64_BITS
uint16_t bpo_end = 16;
#else
uint16_t bpo_end = 8;
#endif

using namespace std;
using namespace v3c;
using namespace treedb;
using namespace meta;

struct Data {
	int32_t m_key;
	char m_data[];
};
struct InputBase {
	const char * m_data;
	int32_t m_key;
};
struct Input : InputBase {
	Input()
	{ m_data = "<null>", m_key = -1; }
	Input(const char * p, int32_t i)
	{ m_data = p, m_key = i; }
	Input & operator=(const Data & d)
	{ m_data = d.m_data; m_key = d.m_key; return *this; }
	inline const char * name() const { return m_data; }
	inline int32_t key() const { return m_key; }
};

enum
{ eApple = 0
, eBanana = 5
, eLemon = 10
, eOrange = 15
, ePear = 20
, ePlum = 25
};
InputBase sorted_items_base[] =
{ { "apple" , eApple }
, { "banana", eBanana }
, { "lemon" , eLemon }
, { "orange", eOrange }
, { "pear"  , ePear }
, { "plum"  , ePlum }
};
InputBase items_base[] =
{ { "apple" , eApple }
, { "orange", eOrange }
, { "lemon" , eLemon }
, { "pear"  , ePear }
, { "plum"  , ePlum }
, { "banana", eBanana }
};
#define NITEMS sizeof(items_base)/sizeof(items_base[0])

Input * sorted_items = (Input *)sorted_items_base;
Input * items = (Input *)items_base;

char buf[TREEDB_PAGE_SIZE]
__attribute__ ((aligned (TREEDB_ALIGN))) = {0};

char * g_buf() { return buf; }

const char * flags_desc(unsigned int flags)
{
	switch(flags) {
	case eTreedbContainerHasCount:
		return "count";
	case eTreedbContainerHasDepth:
		return "depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth:
		return "count | depth";
	case eTreedbContainerHasCount | eTreedbContainerIsPointer:
		return "pointer | count";
	case eTreedbContainerHasDepth | eTreedbContainerIsPointer:
		return "pointer | depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth | eTreedbContainerIsPointer:
		return "pointer | count | depth";
	default:
		return "<unknown flags>";
	}
}

struct StringTreeBackendTypes {
	typedef int32_t KeyType;
	typedef Data DataType;
	typedef Input InputType;
};
struct StringTreeBackend
: public TreeBackendT
			< StringTreeBackend, TreeRuntime, StringTreeBackendTypes
, sizeof(char) >
{
	enum { eKeyOffset = 0 };
	inline static int32_t unpack_key(PCVOID pv)
	{ return ((const Data *)pv)->m_key; }
	inline static const Data & unpack_data(PCVOID pv)
	{ return *(const Data *)pv; }
	inline static Data & unpack_data(PVOID pv)
	{ return *(Data *)pv; }
	inline static const Input & unpack_input(PCVOID pv)
	{ return *(const Input *)pv; }

	inline static HTREEDB_TREEKEY pack_input(const Input & i)
	{ return (HTREEDB_TREEKEY)(&i); }
	inline static HTREEDB_TREEKEY pack_key(const int32_t & k)
	{ return (HTREEDB_TREEKEY)((uintptr_t)(& k) - eKeyOffset); }

	uintptr_t size(TreeRuntime *, const Input & in)
	{ return sizeof(Data) + strlen(in.m_data) + 1; }
	bool assign(TreeRuntime *, Data & to, const Input & from)
	{
		to.m_key = from.m_key;
		strcpy(to.m_data, from.m_data);
		return true;
	}
	inline int compare(int32_t a, int32_t b)
	{
		int ret = (a > b) ? 1 : ( a < b ) ? -1 : 0;
		cout << "StringTreeBackend::compare(" << a << ", " << b << ") == "
			<< ret << '.' << endl;
		return ret;
	}
	inline static char * render(const Data & i)
	{
		char * ret;
		if(asprintf(& ret, "[ %u, %s ]", i.m_key, i.m_data) < 0)
			return 0;
		return ret;
	}
};

/*! A treedb_tree_t delegate.
*/
typedef TreeImplT< StringTreeBackend > StringTreeImpl;
typedef TreeWithImplT< StringTreeBackend > StringTreeWithImpl;

int test(uint16_t abytes, uint16_t aflags, uint16_t xbytes, uint16_t xflags)
{
	cout << "############################################################"
		"####################\n";
	if(aflags & eTreedbContainerIsPointer)
		cout << "# "
#ifdef __cplusplus
		"C++ "
#endif
		"String tree : aos [" << flags_desc(aflags) << "], ";
	else
		cout << "# "
#ifdef __cplusplus
		"C++ "
#endif
		"String tree : aos [" << abytes*8 << " : " << flags_desc(aflags) << "], ";
	if(xflags & eTreedbContainerIsPointer)
		cout << "tree[" << flags_desc(xflags) << "].\n";
	else
		cout << "tree[" << xbytes*8 << " : " << flags_desc(xflags) << "].\n";
	cout << "############################################################"
		"####################" << endl;

	/* The storage. */
	StorageOnMemory storage(g_buf(), 4096);

	/* The aos. */
	AllocatorOnStorageImpl allocator_impl;
	AllocatorOnStorage aos;
	aos.m_inst.m_pImpl = & allocator_impl.m_inst;

	/* Even though only the size member is needed from the header, I'm including
	it to compare the numbers with meta-comet. */
	if(!aos.init_runtime(& allocator_impl.c_type(), abytes, aflags
			, sizeof(TreedbHeader), & storage.base_instance())) {
		cerr << "Error 1: No matching aos template found." << endl;
		return 1;
	}
	treedb_const_tree_ptr_t ptree_template
		= treedb_get_avl_tree_template_for(xbytes, xflags);
	if(!ptree_template) {
		cerr << "Error 2: No matching tree template found." << endl;
		return 2;
	}
	PackedPayload< StringTreeImpl > payload;
	uint64_t payload_64;

	if(!aos.Init
			( ptree_template->m_pImpl->m_pDescription->m_root.m_uSize
			, 4096
			, & payload_64)) {
		cerr << "Error 3: aos.Init() failed." << endl;
		return 3;
	}
	if(payload_64 > (uint64_t)((size_t)-1))
		return 4;
	payload.m_value = (size_t)payload_64;

	cout << "Header: " << sizeof(TreedbHeader)
		<< ", Allocator: " << payload.m_value - sizeof(TreedbHeader)
		<< ", Node: " <<
				aos.free_tree().m_pImpl->m_pDescription->m_node.m_uSize
		<< ", offset: " <<
				aos.free_tree().m_pImpl->m_pDescription->m_pointer.m_uSize
		<< ", Item " << aos.description().m_aligned_item_size
		<< ", Node total " <<
				aos.free_tree().m_pImpl->m_pDescription->m_node.m_uSize
				+ aos.description().m_aligned_item_size
		<< "\n";

	size_t free_offset, free_size;
	if(!aos.LargestFree(&free_offset, &free_size)) {
		cerr << "Error 5: treedb_allocator_LargestFree failed." << endl;
		return 5;
	}

	if(xflags & eTreedbContainerIsPointer)
		payload.m_value = (uintptr_t)aos.OffsetToPointer(payload.m_value);
	StringTreeWithImpl runtime(aos, ptree_template, payload);
	runtime.Init();
	unsigned int n;

	runtime.dump();

	treedb_allocator_t_check_result
		( & aos.base_c_type()
		, aos.Check()
	);
	if(aos.error()) {
		cerr << "Error 6: Allocator check failed : " << aos.error()
			<< "." << endl;
		return 6;
	}
	for(n = 0; n < NITEMS; ++n) {
		cout << "[" << n << "] inserting " << items[n].m_data << "." << endl;
		TreedbInsertResult res = runtime.insert(items[n]);
		if(res != eTreedbInsertSuccess) {
			cerr << "Error 7: [" << n << "] insertion failed: "
				<< TreedbInsertResultDesc(res) << "." << endl;
			return 7;
		}
#if META_TREEDB_DEBUG
		if(!aos.error())
			aos.Check();
#endif
		if(aos.error()) {
			cerr << "Error 8: Allocator check failed : " << aos.error()
				<< "." << endl;
			return 8;
		}
		runtime.dump();
	}
	--n;
	TreedbInsertResult res = runtime.insert(items[n]);
	if(res != eTreedbInsertExists) {
		runtime.dump();
		cerr << "Error 9: [" << n << "] insertion should have failed: "
			<< TreedbInsertResultDesc(res) << "." << endl;
		return 9;
	}
	uintptr_t count = runtime.Count();
	cout << "Inserted " << count << " items.\n";


	PVOID node;
	Input data;

	cout << "search_lt(eAapple)" << endl;
	node = runtime.search_lt(eApple);
	if(node) {
		data = runtime.data(node);
		cout << "Error 10: got \"" << data.m_data << "\" != <null>." << endl;
		return 10;
	}

	cout << "search_lt(ePlum + 1)" << endl;
	node = runtime.search_lt(ePlum + 1);
	if(node) {
		data = runtime.data(node);
		if(data.key() != ePlum) {
			cout << "Error 11: got \"" << data.name() << "\" != \"plum\"."
				<< endl;
			return 11;
		}
	}else{
		cout << "Error 11a: got <null> != \"plum\"." << endl;
		return 11;
	}
	cout << "search_lt(eOrange)" << endl;
	node = runtime.search_lt(eOrange);
	if(node) {
		data = runtime.data(node);
		if(data.key() != eLemon) {
			cout << "Error 12: got \"" << data.name() << "\" != \"lemon\"."
				<< endl;
			return 12;
		}
	}else{
		cout << "Error 12a: got <null> != \"lemon\"." << endl;
		return 12;
	}

	cout << "search_gt(ePlum)" << endl;
	node = runtime.search_gt(ePlum);
	if(node) {
		data = runtime.data(node);
		cout << "Error 13: search_gt(ePlum) : got \"" << data.name()
			<< "\" != <null>." << endl;
		return 13;
	}

	cout << "search_gt(eLemon)" << endl;
	node = runtime.search_gt(eLemon);
	if(node) {
		data = runtime.data(node);
		if(data.key() != eOrange) {
			cout << "Error 14: got \"" << data.name() << "\" != \"orange\"."
				<< endl;
			return 14;
		}
	}else{
		cout << "Error 13a: got <null> != \"orange\"." << endl;
		return 14;
	}

	cout << "search_le(eBanana)" << endl;
	
	if(runtime.search_le(eBanana, node) != eTreedbSearchEmpty) {
		data = runtime.data(node);
		if(data.key() != eBanana) {
			cout << "Error 15: got \"" << data.name() << "\" != \"banana\"."
				<< endl;
			return 15;
		}
	}else{
		cout << "Error 15a: got <null> != \"banana\"." << endl;
		return 15;
	}

	node = runtime.PointerHead();

	for(n = 0; node; ++n, node = runtime.PointerNodeNextPtr(node)) {
		int32_t i = runtime.data(node).m_key;
		const char * p = runtime.data(node).m_data;
		cout << "[" << n << "] " << i << ", " << p
			<< endl;
		if(strcmp(p, sorted_items[n].name())) {
			cerr << "Error 16: [" << n << "] " << p << " != "
				<< sorted_items[n].name() << "." << endl;
			return 16;
		}
		if(i != (int32_t)n*5) {
			cerr << "Error 16a: " << i << " != " << n*5 << "." << endl;
			return 16;
		}
	}

	runtime.Clear();
	treedb_allocator_t_check_result
		( & aos.base_c_type()
		, aos.Check()
	);
	if(aos.error()) {
		cerr << "Error 17: Allocator check failed : " << aos.error()
			<< '.' << endl;
		return 17;
	}

	size_t new_free_offset, new_free_size;
	if(!aos.LargestFree(&new_free_offset, &new_free_size)) {
		cerr << "Error 18: treedb_allocator_LargestFree(2) failed.\n";
		return 18;
	}
	if(free_size != new_free_size) {
		cerr << "Error 19: free size " << new_free_size
			<< " != original free size " << free_size << ".\n";
		return 19;
	}
	if(free_offset != new_free_offset) {
		cerr << "Error 20: free offset " << new_free_offset
			<< " != original free offset " << free_offset << ".\n";
		return 20;
	}
	if(count != NITEMS) {
		cerr << "Error 21: Expected a count of "
		<< NITEMS
		<< ", got " << count << " instead.\n";
		return 21;
	}
	return 0;
}
#define NFLAGS sizeof(flags)/sizeof(flags[0])
#define NPFLAGS sizeof(pflags)/sizeof(pflags[0])
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	unsigned int flags[] =
	{ eTreedbContainerHasCount
	, eTreedbContainerHasDepth
	, eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int pflags[] =
	{ eTreedbContainerIsPointer | eTreedbContainerHasCount
	, eTreedbContainerIsPointer | eTreedbContainerHasDepth
	, eTreedbContainerIsPointer | eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int abytes, aflags_index;
	unsigned int xbytes, xflags_index;
	/* Work through the aos byte range */
	for(abytes = 2; abytes < bpo_end; abytes <<= 1) {
		/* Work through the aos flags */
		for(aflags_index = 0
				; aflags_index < NFLAGS
				; ++aflags_index) {
			/* Work through the tree byte range */
			for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
				/* Work through the tree flags */
				for(xflags_index = 0; xflags_index < NFLAGS; ++xflags_index) {
					int ret = test(abytes, flags[aflags_index], xbytes
							, flags[xflags_index]);
					if(ret)
						return ret;
				}
			}
			/* Work through the pointer tree flags */
			for(xflags_index = 0; xflags_index < NPFLAGS; ++xflags_index) {
				int ret = test(abytes, flags[aflags_index], 0
						, pflags[xflags_index]);
				if(ret)
					return ret;
			}
		}
	}
	/* Work through the pointer aos flags */
	for(aflags_index = 0;
			aflags_index < NPFLAGS
			; ++aflags_index) {
		/* Work through the tree byte range */
		for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
			/* Work through the tree flags */
			for(xflags_index = 0; xflags_index < NFLAGS; ++xflags_index) {
				int ret = test(0, pflags[aflags_index], xbytes
							, flags[xflags_index]);
				if(ret)
					return ret;
			}
		}
		/* Work through the pointer tree flags */
		for(xflags_index = 0;
				xflags_index < NPFLAGS
				; ++xflags_index) {
			int ret = test(0, pflags[aflags_index], 0, pflags[xflags_index]);
			if(ret)
				return ret;
		}
	}
	cout << "Success!\n";
	return 0;
}
