/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/*! \page meta-allocator.c meta-allocator.c
\brief An implementation-independent API for using allocators.

@section allocator-introduction Introduction
The allocators implemented here are independent of the tree
implementations - they just happen to be treedb avl trees here as we have to
start from somewhere.
*/

#include "meta-treedb.h"
#include <malloc.h> /* tree dumps use malloc'd strings :) */

/** Bind the allocator trees to the allocator

Because the treedb_tree_t structures reference an allocator instance, the
allocator contains the meta-data to be used by the allocator trees.

For allocators obtained by treedb_get_allocator_on_storage_template_for()
this is a problem as the structure copy doesn't fix up the pointers.
*/
void treedb_two_tree_allocator_Setup
(treedb_two_tree_allocator_ptr_t p
) {
	p->m_free.m_pImpl->m_pAllocator
		= p->m_used.m_pImpl->m_pAllocator
		= (treedb_allocator_ptr_t)p;
	/* Pointer allocators use real pointers - fixup those handles. */
	if(p->m_pImpl->m_pDescription->m_flags & eTreedbContainerIsPointer) {
		/* The free handles default value is 0 - use the used handle. */
		HTREEDB_HANDLE save = p->m_used.m_handle;
		p->m_used.m_handle = (uintptr_t)treedb_allocator_impl_t_OffsetToPointer(
			(treedb_allocator_impl_ptr_t)(p->m_pImpl), save);
		p->m_free.m_handle += p->m_used.m_handle - save;
	}
}

#define AVL_TMPL_prefix static inline
#define TREEDB_HAS_RESIZABLE_HEAP
#define TREEDB_OMIT_treedb_init
#define TREEDB_OMIT_treedb_init_memory
#define TREEDB_OMIT_treedb_init_header
#define TREEDB_OMIT_pack_unpack_pointer_DECLARE
#define TREEDB_OMIT_treedb_impl_DECLARE
#define TREEDB_OMIT_avl_allocator_impl
#define TREEDB_OMIT_post_free_impl
#define TREEDB_OMIT_treedb_ns_types_typedefs1
#define TREEDB_OMIT_treedb_file_storage_impl

#define TREEDB_NSA(x) meta_treedb_##x
#define meta_treedb_node_ptr_t PVOID
#define meta_treedb_packed_node_ptr_t meta_treedb_count_t
#define meta_treedb_count_t TREEDB_NS_count_t
#define TREEDB_NS_pack_pointer(p) treedb_allocator_t_PointerPack(runtime, p)
#define TREEDB_NS_unpack_pointer(p) treedb_allocator_t_PointerUnpack(runtime, p)

#define TREEDB_node_next(p) treedb_tree_t_PointerNodeNextPtr(& runtime->m_free, p)
#define TREEDB_page_alloc(cb) \
	treedb_storage_t_MoreCore(runtime->m_pImpl->m_pStorage, cb)
#define TREEDB_page_free(cb) \
	treedb_storage_t_FreeCore(runtime->m_pImpl->m_pStorage, cb)
#define TREEDB_post_free(pv, cb) \
	treedb_storage_t_PostFree(runtime->m_pImpl->m_pStorage, pv, cb)

#define TREEDB_PAGE_SIZE 4096
#define TREEDB_ALIGN 8

/* Used tree */
#define meta_treedb_used_impl_unpack_node(ctx, p) \
	treedb_allocator_t_UnpackPointer(ctx, p)

#define meta_treedb_used_search(ctx, t, k) \
	treedb_tree_t_PointerSearch(& ctx->m_used, (PVOID)(uintptr_t)(k))
#define meta_treedb_used_search_closest(ctx, t, k, pn) \
	treedb_tree_t_SearchClosestPtr(& ctx->m_used, (PVOID)(uintptr_t)(k), pn)
#define meta_treedb_used_search_le(ctx, t, k, pn) \
	treedb_tree_t_PointerSearchLe(& ctx->m_used, (PVOID)(uintptr_t)(k), pn)
#define meta_treedb_used_search_ge(ctx, t, k, pn) \
	treedb_tree_t_SearchGePtr(& ctx->m_used, (PVOID)(uintptr_t)(k), pn)
#define meta_treedb_used_unlink_node(ctx, t, p) \
	treedb_tree_t_UnlinkNodePtr(& ctx->m_used, p)
#define meta_treedb_used_insert_node(ctx, t, p) \
	treedb_tree_t_InsertNodePtr(& ctx->m_used, p)

#define meta_treedb_used_count(ctx, t) \
	treedb_tree_t_Count(& ctx->m_used)

/* Free tree */
#define meta_treedb_free_impl_unpack_node(ctx, p) \
	treedb_allocator_t_PointerUnpack(ctx, p)
#define meta_treedb_free_pack_node_offset(u) ((HTREEDB_HANDLE)u)

#define meta_treedb_free_search(ctx, t, k) \
	treedb_tree_t_PointerSearch(& ctx->m_free, (PVOID)(k))
#define meta_treedb_free_search_closest(ctx, t, k, pn) \
	treedb_tree_t_SearchClosestPtr(& ctx->m_free, (PVOID)(k), pn)
#define meta_treedb_free_search_le(ctx, t, k, pn) \
	treedb_tree_t_PointerSearchLe(& ctx->m_free, (PVOID)(k), pn)
#define meta_treedb_free_search_ge(ctx, t, k, pn) \
	treedb_tree_t_SearchGePtr(& ctx->m_free, (PVOID)(k), pn)
#define meta_treedb_free_unlink_node(ctx, t, p) \
	treedb_tree_t_UnlinkNodePtr(& ctx->m_free, p)
#define meta_treedb_free_insert_node(ctx, t, p) \
	treedb_tree_t_InsertNodePtr(& ctx->m_free, p)

#define meta_treedb_free_count(ctx, t) \
	treedb_tree_t_Count(& ctx->m_free)

#define TREEDB_check_heap(runtime) \
	treedb_allocator_t_check_result \
	((treedb_allocator_ptr_t)(runtime), TREEDB_NS(check_heap)(runtime))

#define TREEDB_ALLOC_set_errno(x) \
  treedb_allocator_t_check_result(runtime, x)

#define TREEDB_NS_size_t size_t

#define TREEDB_STORAGE_base \
	(char *)treedb_storage_t_Start(runtime->m_pImpl->m_pStorage)
#define TREEDB_STORAGE_size treedb_storage_t_Size(runtime->m_pImpl->m_pStorage)

#define TREEDB_FUNC0(x) meta_treedb_allocator_##x(runtime)
#define TREEDB_FUNC1(x,y) meta_treedb_allocator_##x(runtime, y)
#define TREEDB_FUNC2(x,y,z) meta_treedb_allocator_##x(runtime, y, z)

#define TREEDB_NS(x) V3C_INVOKE2(V3C_JOIN2,TREEDB_NS_,x)
#define AVL_TMPL_omit_pack_unpack_pointer_DECLARE

#define TREEDB_item_offset(item) (((uintptr_t)item) & ~((uintptr_t)1))
#define TREEDB_item_isfree(item) (((uintptr_t)item) & (uintptr_t)1)
#define TREEDB_item_setfree(item) (((uintptr_t)item) | (uintptr_t)1)

#define treedb_tree_node_Item(p) \
((TREEDB_NS(item_ptr_t)) \
	((char *)(p) + runtime->m_pImpl->m_pDescription->m_aligned_node_size))
#define TREEDB_ALLOC_node_prev(p) \
	treedb_tree_node_Item(p)->m_offset_prev
#define TREEDB_ALLOC_node_set_prev(p, v) \
	treedb_tree_node_Item(p)->m_offset_prev = v
#define TREEDB_ALLOC_node_and_prev(p, v) \
	treedb_tree_node_Item(p)->m_offset_prev &= v
#define TREEDB_ALLOC_node_next(p) \
	treedb_tree_node_Item(p)->m_offset_next
#define TREEDB_ALLOC_node_set_next(p, v) \
	treedb_tree_node_Item(p)->m_offset_next = v
#define TREEDB_ALLOC_node_and_next(p, v) \
	treedb_tree_node_Item(p)->m_offset_next &= v
#define TREEDB_ALLOC_node_length(p) \
	treedb_tree_node_Item(p)->m_uLength
#define TREEDB_ALLOC_node_set_length(p, v) \
	treedb_tree_node_Item(p)->m_uLength = v
#define TREEDB_ALLOC_node_length_add(p, v) \
	treedb_tree_node_Item(p)->m_uLength += v
#define TREEDB_ALLOC_node_length_sub(p, v) \
	treedb_tree_node_Item(p)->m_uLength -= v

#define TREEDB_ALLOC_used_head treedb_tree_t_PointerHead(& runtime->m_used)
#define TREEDB_ALLOC_used_top treedb_tree_t_PointerTop(& runtime->m_used)
#define TREEDB_ALLOC_used_tail treedb_tree_t_PointerTail(& runtime->m_used)

#define TREEDB_ALLOC_free_head treedb_tree_t_PointerHead(& runtime->m_free)
#define TREEDB_ALLOC_free_top treedb_tree_t_PointerTop(& runtime->m_free)
#define TREEDB_ALLOC_free_tail treedb_tree_t_PointerTail(& runtime->m_free)

#define TREEDB_ALLOC_item_size \
	runtime->m_pImpl->m_pDescription->m_aligned_item_size
#define TREEDB_ALLOC_node_size \
	(runtime->m_pImpl->m_pDescription->m_aligned_node_size \
		+ TREEDB_ALLOC_item_size)

#define TREEDB_ALLOC_nodeptr_to_addr(p) ((char *)p + TREEDB_ALLOC_node_size)
#define TREEDB_ALLOC_addr_to_nodeptr(p) \
	((TREEDB_NS(item_ptr_t))((char *)p - TREEDB_ALLOC_node_size))

#define TREEDB_ALLOC_used_tree \
  (TREEDB_STORAGE_base + runtime->m_used.m_handle)
#define TREEDB_ALLOC_free_tree \
  (TREEDB_STORAGE_base + runtime->m_free.m_handle)

/* Used in meta-allocator-impl.h */
#define treedb_tree_impl_node_Item(p) \
((TREEDB_NS(item_ptr_t)) \
	((char *)(p) + This->m_pDescription->m_aligned_node_size))
#define META_TREEDB_ALLOC_node_prev(p) \
	treedb_tree_impl_node_Item(p)->m_offset_prev
#define META_TREEDB_ALLOC_node_set_prev(p, v) \
	treedb_tree_impl_node_Item(p)->m_offset_prev = v
#define META_TREEDB_ALLOC_node_and_prev(p, v) \
	treedb_tree_impl_node_Item(p)->m_offset_prev &= v
#define META_TREEDB_ALLOC_node_next(p) \
	treedb_tree_impl_node_Item(p)->m_offset_next
#define META_TREEDB_ALLOC_node_set_next(p, v) \
	treedb_tree_impl_node_Item(p)->m_offset_next = v
#define META_TREEDB_ALLOC_node_and_next(p, v) \
	treedb_tree_impl_node_Item(p)->m_offset_next &= v
#define META_TREEDB_ALLOC_node_length(p) \
	treedb_tree_impl_node_Item(p)->m_uLength
#define META_TREEDB_ALLOC_node_set_length(p, v) \
	treedb_tree_impl_node_Item(p)->m_uLength = v
#define META_TREEDB_ALLOC_node_length_add(p, v) \
	treedb_tree_impl_node_Item(p)->m_uLength += v
#define META_TREEDB_ALLOC_node_length_sub(p, v) \
	treedb_tree_impl_node_Item(p)->m_uLength -= v
#define META_TREEDB_ALLOC_item_size This->m_pDescription->m_aligned_item_size
#define META_TREEDB_ALLOC_node_size \
	(This->m_pDescription->m_aligned_node_size + META_TREEDB_ALLOC_item_size)
#define META_TREEDB_ALLOC_nodeptr_to_addr(p) \
	((char *)p + META_TREEDB_ALLOC_node_size)
#define META_TREEDB_ALLOC_addr_to_nodeptr(p) \
	((TREEDB_NS(item_ptr_t))((char *)p - META_TREEDB_ALLOC_node_size))
#define META_TREEDB_ALLOC_used_tree (pAllocator->m_used.m_handle)
#define META_TREEDB_ALLOC_free_tree (pAllocator->m_free.m_handle)

#define TREEDB_pack_pointer(p) \
	treedb_allocator_t_PointerPack(runtime, p)
#define TREEDB_unpack_pointer(p) \
	treedb_allocator_t_PointerUnpack(runtime, p)

#define TREEDB_pointer_to_offset(p) \
	treedb_allocator_t_PointerToOffset(runtime, p)
#define TREEDB_offset_to_pointer(p) \
	treedb_allocator_t_OffsetToPointer(runtime, p)

#define META_TREEDB_pack_pointer(p) \
	treedb_allocator_impl_t_PointerPack(This, p)
#define META_TREEDB_unpack_pointer(p) \
	treedb_allocator_impl_t_PointerUnpack(This, p)

#define META_TREEDB_pointer_to_offset(p) \
	treedb_allocator_impl_t_PointerToOffset(This, p)
#define META_TREEDB_offset_to_pointer(p) \
	treedb_allocator_impl_t_OffsetToPointer(This, p)

/* This is needed for meta-allocator.h */
#define AVL_TMPL_NS(x) TREEDB_NS(x)
#define AVL_TMPL_count_t TREEDB_NS_count_t

/******************************************************************************/
/* o16                                                                        */
/******************************************************************************/
#define TREEDB_NS_ treedb_allocator_o16_
#define TREEDB_NBITS 16

#define TREEDB_NS_count_t uint16_t
#define TREEDB_NS_pack_t treedb_packed_uint16_t

#ifdef DOXYGEN_processing
{
#endif

#include "meta-allocator-impl.h"

#ifdef DOXYGEN_processing
}
#endif

/******************************************************************************/
/* o32                                                                        */
/******************************************************************************/
#undef TREEDB_NS_
#undef TREEDB_NS_count_t
#undef TREEDB_NS_pack_t
#undef TREEDB_NBITS
#define TREEDB_NBITS 32
#define TREEDB_NS_ treedb_allocator_o32_
#define TREEDB_NS_count_t uint32_t
#define TREEDB_NS_pack_t treedb_packed_uint32_t

#ifdef DOXYGEN_processing
{
#endif

#include "meta-allocator-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#if HAVE_64_BITS
/******************************************************************************/
/* o64                                                                        */
/******************************************************************************/
#undef TREEDB_NS_
#undef TREEDB_NS_count_t
#undef TREEDB_NS_pack_t
#undef TREEDB_NBITS
#define TREEDB_NBITS 64
#define TREEDB_NS_ treedb_allocator_o64_
#define TREEDB_NS_count_t uint64_t
#define TREEDB_NS_pack_t treedb_packed_uint64_t

#ifdef DOXYGEN_processing
{
#endif

#include "meta-allocator-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#endif /* #if HAVE_64_BITS */

/******************************************************************************/
/* pointer                                                                    */
/******************************************************************************/
#undef TREEDB_NS_
#undef TREEDB_NS_count_t
#undef TREEDB_NS_pack_t
#undef TREEDB_NBITS
#define TREEDB_NBITS 0

#undef TREEDB_unpack_pointer
#undef TREEDB_pack_pointer
#define TREEDB_unpack_pointer(x) ((PVOID)(x))
#define TREEDB_pack_pointer(x) ((TREEDB_NS_count_t)(x))

#define v3c_nbits_u0(x) v3c_nbits_up(x)
#define uint0_t uintptr_t

#define TREEDB_NS_ treedb_allocator_pointer_
#define TREEDB_NS_count_t uintptr_t
#define TREEDB_NS_pack_t treedb_packed_uintptr_t
#undef meta_treedb_used_impl_unpack_node
#undef meta_treedb_free_impl_unpack_node
#undef TREEDB_NS_max_offset_value
#define meta_treedb_used_impl_unpack_node(runtime, p) ((PVOID)p)
#define meta_treedb_free_impl_unpack_node(runtime, p) ((PVOID)p)
#define TREEDB_NS_offset_out_of_range(x) 0

#ifdef DOXYGEN_processing
{
#endif

#include "meta-allocator-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef TREEDB_item_offset
#undef TREEDB_item_isfree
#undef TREEDB_item_setfree
