/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifndef v3c_meta_treedb_tree_hpp_included
  #define v3c_meta_treedb_tree_hpp_included

  #include <v3c/trace_scope.hpp>

  #ifndef v3c_meta_treedb_tree_h_included
	#define COBJMACROS
	#include "meta-treedb-tree.h"
  #endif

  #include <iostream>

namespace v3c {
namespace treedb {
namespace meta {

using namespace std;

class TreeRuntime;
/*! treedb_tree_impl_t vtable-only overlay

Since C++ doesn't have static virtual functions, this is done the "C" way.

Remember, the first parameter is a TreeRuntime *, from which we access the
m_pImpl member to get at "our" m_pVtbl - not "this".

It's a mess, but it's the only way I can figure to pass a "real" runtime to the
backend.

The "real" runtime is the client-side Tree instance, or anything with a TreeImpl
pointer as its first data member, not a TreeImpl.

This avoids having to store client runtime in the backend - it shouldn't be
tied to any of its clients.

This is only for C++ convenience, which is funny as we can't ever implement a
TreeImpl using C++ virtual functions.
*/
class ITreeImpl_VtableOverlay
{
public:
#include "meta-treedb-cxx-push-c.h"
	treedb_tree_impl_t_methods(,,TreeRuntime);
#include "meta-treedb-cxx-pop-c.h"
	treedb_tree_impl_vtbl_ptr_t m_pVtbl;
protected:
	//! This can't be virtual. If it were it would mess up derived interfaces.
	~ITreeImpl_VtableOverlay() {}
};

/*! treedb_tree_impl_t overlay
*/
template< typename Interface = ITreeImpl_VtableOverlay >
class ITreeImplT : public Interface
{
public:
	inline Allocator & allocator()
	{ return *(Allocator *)(m_pAllocator); }
	inline const TreedbItemDescription & description() const
	{ return * m_pDescription; }
	inline char * node_to_data(PVOID pn)
	{ return (char *)pn + description().m_aligned_node_size; }
	inline char * data_to_node(PVOID pv)
	{ return (char *)pv - description().m_aligned_node_size; }

	treedb_tree_impl_t_members(treedb_tree_impl_vtbl_ptr_t);
};

/*! treedb_tree_renderer_t overlay

Because treedb_tree_renderer_t doesn't have data members (besides the vtable)
the step of defining a VtableOverlay is skipped here.
*/
class ITreeRenderer
{
public:
	treedb_tree_renderer_t_methods(,,);

	inline treedb_tree_renderer_t & c_type()
	{ return *(treedb_tree_renderer_ptr_t)this; }
protected:
	//! OK this could be virtual.
	~ITreeRenderer() {}
};

/*! treedb_tree_renderer_t overlay
*/
class ITreeDataRenderer
{
public:
	treedb_tree_data_renderer_t_methods(,,);

	treedb_tree_data_renderer_t & c_type()
	{ return  *(treedb_tree_data_renderer_ptr_t)this; }
protected:
	//! OK this could be virtual.
	~ITreeDataRenderer() {}
};

/*! treedb_tree_backend_t overlay

Because treedb_tree_backend_t doesn't have data members (besides the vtable)
the step of defining a VtableOverlay is skipped here.
*/
template< typename RuntimeType >
class ITreeBackendT
{
public:
	treedb_tree_backend_t_methods(,,,RuntimeType);
protected:
	//! OK this could be virtual.
	~ITreeBackendT() {}
};

////////////////////////////////////////////////////////////////////////////////
// Wrappers
////////////////////////////////////////////////////////////////////////////////
#define DGTHIS_MEM(x,m) x.m
#define DGTHIS0(x) treedb_tree_t_to_runtime(& (x))
#define DGTHIS(x) treedb_tree_t_to_runtime(& (x)),
#define V3C_DELEGATE(Inst, x) Inst.m_pImpl->m_pVtbl->x

class TreeImpl;

/*! treedb_tree_t wrapper

This class CANNOT have a vtable! It overlays it's m_inst member.
*/
class Tree
{
public:
	inline treedb_tree_t & instance()
	{ return m_inst; }
	inline const treedb_tree_t & instance() const
	{ return m_inst; }

	/* Convenience accessors. */
	inline const treedb_tree_impl_t & impl() const
	{ return * m_inst.m_pImpl; }
	inline const TreedbItemDescription & description() const
	{ return * impl().m_pDescription; }
	inline Allocator & allocator()
	{ return *(Allocator *)(impl().m_pAllocator); }

	inline char * node_to_data(PVOID pn)
	{ return (char *)pn + description().m_aligned_node_size; }
	inline char * data_to_node(PVOID pv)
	{ return (char *)pv - description().m_aligned_node_size; }

	inline HTREEDB_TREE raw_handle() const { return m_inst.m_handle; }
	inline HTREEDB_TREE & raw_handle() { return m_inst.m_handle; }

	TREEDB_ITree_DELEGATE_TO(m_inst);
	treedb_tree_t m_inst;
};

#undef V3C_DELEGATE
#undef DGTHIS
#undef DGTHIS0

class TreeImpl;

/*! treedb_tree_runtime_t wrapper

This class CANNOT have a vtable! It overlays it's m_inst member.
*/
class TreeRuntime
{
public:
	inline TreeRuntime() { m_inst.m_pImpl = 0; }
	inline TreeRuntime(TreeImpl * p)
		{ m_inst.m_pImpl = (treedb_tree_impl_ptr_t)p; }
	/* Convenience accessors. */
	inline const TreedbItemDescription & description() const
	{ return * m_inst.m_pImpl->m_pDescription; }
	inline Allocator & allocator()
	{ return *(Allocator *)(m_inst.m_pImpl->m_pAllocator); }

	inline char * node_to_data(PCVOID pn)
	{ return (char *)pn + description().m_aligned_node_size; }
	inline char * data_to_node(PCVOID pv)
	{ return (char *)pv - description().m_aligned_node_size; }
	inline TreeImpl & base_impl();

	treedb_tree_runtime_t m_inst;
};

#define DGTHIS0(x) (&x)
#define DGTHIS(x) (&x),
#define V3C_DELEGATE(Inst, x) Inst.m_pVtbl->x

/*! treedb_tree_impl_t wrapper

This class CANNOT have a vtable! It overlays it's m_inst member.
*/
class TreeImpl
{
public:
	inline TreeImpl() : m_inst() {}
	inline TreeImpl
	( Allocator & allocator
	, treedb_tree_backend_ptr_t pBackend
	, TreedbItemDescription * pd
	, treedb_const_tree_impl_ptr_t ptemplate
	);
	inline void setup
	( Allocator & allocator
	, treedb_tree_backend_ptr_t pBackend
	, TreedbItemDescription * pd
	, treedb_const_tree_impl_ptr_t ptemplate
	);
	inline void setup
	( AllocatorOnStorage & allocator
	, treedb_tree_backend_ptr_t pBackend
	, TreedbItemDescription * pd
	, treedb_const_tree_impl_ptr_t ptemplate
	);
	/* Convenience accessors. */
	inline const TreedbItemDescription & description() const
	{ return * m_inst.m_pDescription; }
	inline Allocator & allocator()
	{ return *(Allocator *)(m_inst.m_pAllocator); }

	inline char * node_to_data(PCVOID pn)
	{ return (char *)pn + description().m_aligned_node_size; }
	inline char * data_to_node(PCVOID pv)
	{ return (char *)pv - description().m_aligned_node_size; }

	TREEDB_ITreeImpl_DELEGATE_TO(m_inst);
	treedb_tree_impl_t m_inst;
};

inline TreeImpl & TreeRuntime::base_impl()
	{ return *(TreeImpl *)m_inst.m_pImpl; }

/*! treedb_tree_backend_t wrapper

This class CANNOT have a vtable! It overlays it's m_inst member.
*/
class TreeBackend
{
public:
	TREEDB_ITreeBackend_DELEGATE_TO(m_inst);
	treedb_tree_backend_t m_inst;
};

/*! treedb_fd_tree_renderer_t wrapper

This class CANNOT have a vtable! It overlays it's m_inst member.
*/
class FdTreeRenderer
{
public:
	FdTreeRenderer(int fd)
	{
		static treedb_fd_tree_renderer_t ref = INIT_TREEDB_FD_TREE_RENDERER(0);
		m_inst = ref;
		m_inst.m_printer.fd = fd;
	}
	inline treedb_tree_renderer_t & c_type()
	{ return *(treedb_tree_renderer_ptr_t)(& m_inst); }
	inline operator treedb_tree_renderer_ptr_t() { return & c_type(); }

	inline operator ITreeRenderer &()
	{ return *(ITreeRenderer *)(& m_inst); }
	~FdTreeRenderer()
	{
		v3c_tree_printer_term(& m_inst.m_printer);
	}
	TREEDB_ITreeRenderer_DELEGATE_TO(m_inst);
	treedb_fd_tree_renderer_t m_inst;
};

#undef V3C_DELEGATE
#undef DGTHIS0
#undef DGTHIS
#undef DGTHIS_MEM

////////////////////////////////////////////////////////////////////////////////
// Templates
////////////////////////////////////////////////////////////////////////////////
template< typename Impl >
class TreeRuntimeT
: public TreeRuntime
{
public:
	inline TreeRuntimeT() : TreeRuntime() {}
	inline TreeRuntimeT(Impl * p) : TreeRuntime((Impl *)p) {}
	inline Impl & impl() { return *(Impl *)m_inst.m_pImpl; }
};

template< typename Backend_ >
class TreeImplT
: public TreeImpl
{
public:
	typedef Backend_ Backend;
	typedef typename Backend::KeyType KeyType;
	typedef typename Backend::DataType DataType;
	typedef typename Backend::InputType InputType;

	inline void setup
	( Allocator & allocator, treedb_const_tree_impl_ptr_t ptemplate)
	{
		TreeImpl::setup(allocator, (treedb_tree_backend_ptr_t)(& m_backend)
			, & m_description, ptemplate);
	}
	inline void setup
	( AllocatorOnStorage & allocator, treedb_const_tree_impl_ptr_t ptemplate)
	{
		TreeImpl::setup(allocator, (treedb_tree_backend_ptr_t)(& m_backend)
			, & m_description, ptemplate);
	}
	inline TreeImplT() {}
	inline TreeImplT
	( Allocator & allocator
	, treedb_const_tree_impl_ptr_t ptemplate
	)
	: TreeImpl()
	{
		setup(allocator, ptemplate);
	}
	inline TreeImplT
	( AllocatorOnStorage & allocator
	, treedb_const_tree_impl_ptr_t ptemplate
	)
	: TreeImpl()
	{
		setup(allocator, ptemplate);
	}
	// Accessors
	KeyType key(PCVOID pv)
	{
		return Backend::unpack_key(node_to_data(pv));
	}
	const DataType & data(PCVOID pv)
	{
		return Backend::unpack_data(node_to_data(pv));
	}
	DataType & data(PVOID pv)
	{
		return Backend::unpack_data(node_to_data(pv));
	}
#if 0
	inline bool remove_many
	( HTREEDB_TREE ht
	, KeyType lower
	, KeyType upper
	, TreedbKeyDistanceFlags flags
	) { return RemoveMany(ht, backend().pack_key(lower)
		, backend().pack_key(upper), flags);
	}
	inline bool key_distance
	( HTREEDB_TREE ht
	, KeyType lower
	, KeyType upper
	, TreedbKeyDistanceFlags flags
	, intptr_t * pres
	) { return KeyDistance(ht, backend().pack_key(lower)
		, backend().pack_key(upper), flags, pres);
	}
	inline TreedbInsertResult insert
	( HTREEDB_TREE ht
	, InputType i
	, PVOID * ppn = 0
	) { return TreeImpl::Insert(ht, backend().pack_input(i), ppn); }
	inline bool remove
	( HTREEDB_TREE ht
	, KeyType i
	) { return TreeImpl::Remove(ht, backend().pack_key(i)); }
	inline bool contains(KeyType i)
	{ return Tree::Search(backend().pack_key(i)) ? true : false; }
#endif
	inline Backend & backend() { return m_backend; }
	Backend m_backend;
	TreedbItemDescription m_description;
};

template < typename Backend_, typename Impl_ = TreeImplT< Backend_ > >
class TreeWithImplT;

template < typename Impl_ >
class TreeT : public Tree
{
public:
	typedef Impl_ Impl;
	typedef TreeT< Impl > TreeType;
	typedef typename Impl::Backend Backend;
	typedef typename Backend::KeyType KeyType;
	typedef typename Backend::DataType DataType;
	typedef typename Backend::InputType InputType;
	typedef PackedPayload< Impl > HandleType;

	inline HandleType handle() const
		{ return HandleType(m_inst.m_handle); }
	inline HandleType & handle()
		{ return * (HandleType *)(& m_inst.m_handle); }

	TreeT() {}
	inline void setup(Impl_ & timpl, HandleType payload)
	{ treedb_tree_t_Setup(& m_inst, & timpl.m_inst, payload.m_value); }
	inline TreeT(Impl & timpl, HandleType payload)
	{ this->setup(timpl, payload); }
	TreeT(TreeWithImplT< Backend, Impl > & o, HandleType payload);
	inline Impl & impl() { return (Impl &)Tree::impl(); }
	inline Backend & backend() { return impl().backend(); }
	inline bool remove_many
	( KeyType lower
	, KeyType upper
	, TreedbKeyDistanceFlags flags
	) { return Tree::RemoveMany(backend().pack_key(lower)
		, backend().pack_key(upper), flags);
	}
	inline bool key_distance
	( KeyType lower
	, KeyType upper
	, TreedbKeyDistanceFlags flags
	, intptr_t * pres
	) { return Tree::KeyDistance(backend().pack_key(lower)
		, backend().pack_key(upper), flags, pres);
	}
	inline TreedbInsertResult insert(InputType i, PVOID * ppn = 0)
	{ return Tree::Insert(backend().pack_input(i), ppn); }
	inline bool remove(KeyType i)
	{ return Tree::Remove(backend().pack_key(i)); }
	inline bool contains(KeyType i)
	{ return Tree::Search(backend().pack_key(i)) ? true : false; }
	void dump();
	inline bool clone(TreeT< Impl_ > & from)
	{ return Tree::Clone(from.m_inst.m_handle); }

	inline PVOID search(KeyType i)
	{ return Tree::PointerSearch(impl().backend().pack_key(i)); }
	inline PVOID search_lt(KeyType i)
	{ return Tree::PointerSearchLt(impl().backend().pack_key(i)); }
	inline PVOID search_gt(KeyType i)
	{ return Tree::PointerSearchGt(impl().backend().pack_key(i)); }
	inline TreedbSearchResult search_le(KeyType i, PVOID & pn)
	{ return Tree::SearchLePtr(impl().backend().pack_key(i), & pn); }
	inline TreedbSearchResult search_ge(KeyType i, PVOID & pn)
	{ return Tree::SearchGePtr(impl().backend().pack_key(i), & pn); }

	class DataRenderer : public ITreeDataRenderer
	{
	public:
		DataRenderer(uint16_t aligned_node_size)
		: m_aligned_node_size(aligned_node_size)
		{}
		virtual PSTR STDCALL Dump
		( treedb_tree_runtime_ptr_t runtime
		, PCVOID pv
		) {
			(void)runtime;
			char * p = ((char *)pv) + m_aligned_node_size;
	#if META_TREEDB_DEBUG
			cout << __FILE__ << '[' << __LINE__ << "] TreeT::DataRenderer::Dump"
				" : Reading \""
				<< Impl::Backend::unpack_data(p) << "\" from "
				<< runtime->m_pImpl->m_pAllocator->PointerToOffset(pv)
				<< " + " << m_aligned_node_size
				<< " = "
				<< runtime->m_pImpl->m_pAllocator->PointerToOffset(p)
				<< "." << endl;
	#endif
			char * ret = Impl::Backend::render(Impl::Backend::unpack_data(p));
			return ret;
		}
		uint16_t m_aligned_node_size;
	};

	// Accessors
	const DataType & data(PCVOID pv)
	{
		return Impl::Backend::unpack_data(node_to_data(pv));
	}
	DataType & data(PVOID pv)
	{
		return Impl::Backend::unpack_data(node_to_data(pv));
	}
};

template < typename Backend_, typename Impl_ /*= TreeImplT< Backend_ >*/ >
class TreeWithImplT : public TreeT< Impl_ >
{
public:
	typedef Backend_ Backend;
	typedef Impl_ Impl;
	typedef typename Backend::KeyType KeyType;
	typedef TreeT< Impl > TreeType;
	inline TreeWithImplT()
	: TreeType()
	, m_impl()
	{}
	inline const TreeType & base() const { return *this; }
	inline const TreeType & const_base() const { return *this; }
	inline TreeType & base() { return *this; }
	inline TreeWithImplT(
		Allocator & allocator, treedb_const_tree_ptr_t ptemplate
		, PackedPayload< Impl > payload)
	: TreeType()
	, m_impl(allocator, ptemplate->m_pImpl)
	{ TreeType::setup(m_impl, payload); }
	inline TreeWithImplT(
		AllocatorOnStorage & allocator, treedb_const_tree_ptr_t ptemplate
		, PackedPayload< Impl > payload)
	: TreeType()
	, m_impl(allocator, ptemplate->m_pImpl)
	{ TreeType::setup(m_impl, payload); }
	inline void setup
	( Allocator & allocator
	, treedb_const_tree_ptr_t ptemplate
	, PackedPayload< Impl > payload
	) {
		m_impl.setup(allocator, ptemplate->m_pImpl);
		TreeType::setup(m_impl, payload);
	}
	inline void setup
	( AllocatorOnStorage & allocator
	, treedb_const_tree_ptr_t ptemplate
	, PackedPayload< Impl > payload
	) {
		m_impl.setup(allocator, ptemplate->m_pImpl);
		TreeType::setup(m_impl, payload);
	}
	Impl m_impl;
};

template < typename Impl >
TreeT< Impl >::TreeT(TreeWithImplT< TreeT< Impl >::Backend, Impl > & o
, PackedPayload< Impl > payload
) {
	this->setup(o.impl(), & o.m_inst, payload);
}

template
< typename Derived
, typename Runtime
, typename BackendTypes
, uint16_t align
>
class TreeBackendT : public ITreeBackendT< Runtime >
{
public:
	typedef typename BackendTypes::KeyType KeyType;
	typedef typename BackendTypes::DataType DataType;
	typedef typename BackendTypes::InputType InputType;

	virtual HTREEDB_TREEKEY STDCALL UnpackKey
	( Runtime * runtime
	, PCVOID pv
	) {
		return runtime->node_to_data(pv);
	}
	virtual int STDCALL Compare
	( Runtime * runtime
	, HTREEDB_TREEKEY av
	, HTREEDB_TREEKEY bv
	) {
		TraceWrapX(__PRETTY_FUNCTION__);
		(void)runtime;
		Derived * This = static_cast<Derived *>(this);
		KeyType a = This->unpack_key(av);
		KeyType b = This->unpack_key(bv);
		int res = This->compare(a, b);
	#if META_TREEDB_DEBUG
		bool ain = runtime->allocator().InHeapPtr(av);
		bool bin = runtime->allocator().InHeapPtr(bv);
		trace << "TreeBackendT::Compare \""
			<< a << "\" [";
		if(ain) {
			trace << runtime->allocator().PointerToOffset(av);
		}else{
			trace << "external";
		}
		trace << "] with \"" << b << "\" [";
		if(bin) {
			trace << runtime->allocator().PointerToOffset(bv);
		}else{
			trace << "external";
		}
		trace << "] : " << res << '.' << endl;
	#endif
		return res;
	}
	virtual PVOID STDCALL AllocNode
	( Runtime * runtime
	, HTREEDB_TREEKEY input
	) {
		Derived * This = static_cast<Derived *>(this);
		uintptr_t offset = runtime->description().m_aligned_node_size;
		uintptr_t data_size = This->size(runtime, Derived::unpack_input(input));
	#if META_TREEDB_DEBUG
		trace << __FILE__ << '[' << __LINE__ << "] TreeBackendT::AllocNode : "
			"got data size " << data_size << '.' << endl;
	#endif
		char * ret = (char *)runtime->allocator().PointerAlignedAlloc
					(offset + data_size, align, offset);
		if(!ret)
			return 0;
		char *p = ret + offset;
		InputType in = Derived::unpack_input(input);
	#if META_TREEDB_DEBUG
		trace << __FILE__ << '[' << __LINE__ << "] TreeBackendT::AllocNode : "
			"writing \"" << in << "\" to "
			<< runtime->allocator().PointerToOffset(ret)
			<< " + " << offset << " == "
			<< runtime->allocator().PointerToOffset(p)
			<< "." << endl;
	#endif
		if(!This->assign(runtime, Derived::unpack_data(p), in)) {
			runtime->allocator().FreePtr(ret);
			return 0;
		}
		return ret;
	}
	virtual void STDCALL FreeNode
	( Runtime * runtime
	, PVOID pn
	) {
		runtime->allocator().FreePtr(pn);
	}
	// Default implementations
	inline static char * render(const DataType &)
	{ return 0; }
};
template
< typename Derived
, typename Runtime
, typename BackendTypes
, uint16_t align
>
class SimpleTreeBackendT
: public TreeBackendT< Derived, Runtime, BackendTypes, align >
{
public:
	typedef typename BackendTypes::KeyType KeyType;
	typedef typename BackendTypes::DataType DataType;
	typedef typename BackendTypes::InputType InputType;

	// data == input == key (for demos)
	inline static const KeyType & unpack_key(PCVOID pv)
	{ return *(KeyType *)(&Derived::unpack_data(pv)); }
	inline static const InputType & unpack_input(PCVOID pv)
	{ return *(InputType *)(&Derived::unpack_data(pv)); }
	inline static HTREEDB_TREEKEY pack_key(const KeyType & k)
	{ return Derived::pack_input(k); }
};
template
< typename Derived
, typename Runtime
, typename BackendTypes
, uint16_t align
>
class FixedTreeBackendT : public TreeBackendT< Derived, Runtime, BackendTypes
, align >
{
public:
	typedef typename BackendTypes::KeyType KeyType;
	typedef typename BackendTypes::DataType DataType;
	typedef typename BackendTypes::InputType InputType;

	uintptr_t size(Runtime *, const InputType &)
		{ return sizeof(DataType); }
	bool assign(Runtime *, DataType & to, const InputType & from)
	{
		to = from;
		return true;
	}
};
template
< typename Derived
, typename Runtime
, typename BackendTypes
, uint16_t align
>
class FixedSimpleTreeBackendT
: public SimpleTreeBackendT< Derived, Runtime, BackendTypes, align >
{
public:
	typedef typename BackendTypes::KeyType KeyType;
	typedef typename BackendTypes::DataType DataType;
	typedef typename BackendTypes::InputType InputType;

	uintptr_t size(Runtime *, const InputType &) { return sizeof(DataType); }
	bool assign(Runtime *, DataType & to, const InputType & from)
	{
		to = from;
		return true;
	}
};

} // namespace meta
} // namespace treedb
} // namespace v3c

#endif /* #ifndef v3c_meta_treedb_tree_hpp_included */
