/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/*! \page meta-treedb.c meta-treedb.c
\brief An implementation-independent API for using treedb trees.

@section Introduction Introduction
In order to create, destroy and manipulate treedb trees one needs an
implementation.

Because we cannot instantiate classes by GUID (this is not COM), meta-treedb
provides an API to instantiate some well-known implementations, e.g. AVL.

With this done, one still needs access to some of the fundamental structural
information about the tree
 -# the size of a pointer when stored in the tree
	This is crucial as they may not be stored "as is", but as an offset.
 -# the size and alignment requirements of a tree "root"
 -# the size and alignment requirements of a tree "node"
 -# the aliged size of a tree node

These items also have alignment restrictions, as may any other items you choose
to store after them.

The TreedbTreeDescription and TreedbTreeDescriptionEntry structures contain
this data.

One also needs
 -# A means to allocate storage
 -# A means to compare keys
 -# The tree implementation

Because this information is constant regardless of the tree instance, it is
stored in the treedb_tree_t structure, and almost all trees need a fully
populated one in order to be useful.

The allocator trees don't use the allocator for storage, just packing/unpacking.

*/
#include <stdlib.h>
#include <string.h>
#include "meta-treedb.h"

#ifdef __cplusplus
extern "C" {
#endif

/** Compute the entries in p, referencing the supplied schema if required
*/
ETreedbComputeStructOffsetResult
treedb_compute_struct_offsets
( TreedbItemDescriptionEntryConstPtr p
, TreedbItemDescriptionEntryConstPtr schema
, uintptr_t * puStart
, uintptr_t * out
) {
	uintptr_t uStart = * puStart;
	while(p->m_uSize) {
		if(!p->m_uAlign)
			return eTreedbComputeStructOffsetZeroAlign;
		if(p->m_uSize < 0) {
			uStart += schema->m_uAlign - 1;
			uStart &= ~(schema->m_uAlign - 1);
			*out++ = uStart;
			uStart += (-p->m_uSize) * schema->m_uSize;
		}else{
			if(p->m_uSize & (p->m_uAlign - 1))
				return eTreedbComputeStructOffsetBadSize;
			uStart += schema->m_uAlign - 1;
			uStart &= ~(schema->m_uAlign - 1);
			*out++ = uStart;
			uStart += p->m_uSize;
		}
		++p;
	}
	* puStart = uStart;
	return eTreedbComputeStructOffsetSuccess;
}

/******************************************************************************/
/* treedb_storage_on_memory                                                   */
/******************************************************************************/
static TreedbHeapStatus STDCALL mem_ReadLock(treedb_storage_on_memory_ptr_t This)
{
	(void)This;
	return eTreedbHeapNotOpen;
}
static TreedbHeapStatus STDCALL mem_WriteLock(treedb_storage_on_memory_ptr_t This)
{
	(void)This;
	return eTreedbHeapNotOpen;
}
static TreedbHeapStatus STDCALL mem_Unlock(treedb_storage_on_memory_ptr_t This)
{
	(void)This;
	return eTreedbHeapNotOpen;
}
static StorageLockType STDCALL mem_LockType(treedb_storage_on_memory_ptr_t This)
{
	(void)This;
	return eStorageUnlocked;
}
static BOOL STDCALL mem_IsOpen(treedb_storage_on_memory_ptr_t This)
{
	(void)This;
	return 0;
}
static BOOL STDCALL mem_CanWrite(treedb_storage_on_memory_ptr_t This)
{
	(void)This;
	return 0;
}
static BOOL STDCALL mem_CanExec(treedb_storage_on_memory_ptr_t This)
{
	(void)This;
	return 0;
}
static TreedbHeapStatus STDCALL mem_Close(treedb_storage_on_memory_ptr_t This)
{
	(void)This;
	return eTreedbHeapNotOpen;
}
static TreedbHeapStatus STDCALL mem_OpenUnknown
( treedb_storage_on_memory_ptr_t This
, const char * name
, TreedbHeapOpenFlags open_flags
, int access_flags
, uintptr_t offset
) {
	(void)This;
	(void)name;
	(void)open_flags;
	(void)access_flags;
	(void)offset;
	return eTreedbHeapNotOpen;
}
static TreedbHeapStatus STDCALL mem_Open
( treedb_storage_on_memory_ptr_t This
, const uuid_t uSchema
, const char * name
, TreedbHeapOpenFlags open_flags
, int access_flags
, size_t offset
) {
	(void)This;
	(void)uSchema;
	(void)name;
	(void)open_flags;
	(void)access_flags;
	(void)offset;
	return eTreedbHeapNotOpen;
}
static TreedbHeapStatus STDCALL mem_Create
( treedb_storage_on_memory_ptr_t This
, treedb_allocator_ptr_t pAllocator
, const uuid_t uSchema
, const char * name
, TreedbHeapOpenFlags open_flags
, int create_permissions
, int access_flags
, size_t offset
, size_t size
, size_t uPayloadSize
) {
	(void)This;
	(void)pAllocator;
	(void)uSchema;
	(void)name;
	(void)open_flags;
	(void)create_permissions;
	(void)access_flags;
	(void)offset;
	(void)size;
	(void)uPayloadSize;
	return eTreedbHeapNotOpen;
}
static PVOID STDCALL mem_MoreCore(treedb_storage_on_memory_ptr_t This, size_t cb)
{
	(void)This;
	(void)cb;
	return 0;
}
static TreedbFreeResult STDCALL mem_FreeCore(treedb_storage_on_memory_ptr_t This
, size_t cb
) {
	(void)This;
	(void)cb;
	return eTreedbFreeHeapIsFixed;
}
static BOOL STDCALL mem_Resizable(treedb_storage_on_memory_ptr_t This)
{
	(void)This;
	return 0;
}
static void STDCALL mem_PostFree(treedb_storage_on_memory_ptr_t This, PVOID pv
, size_t cb)
{
	(void)This;
	(void)pv;
	(void)cb;
}
static uint8_t STDCALL mem_EndianIndex
( treedb_storage_on_memory_ptr_t This
) {
	(void)This;
	return v3c_native_endian_index();
}
static uint8_t STDCALL mem_EndianSwapIndex
( treedb_storage_on_memory_ptr_t This
) {
	(void)This;
	return 0; // Native.
}
#if 0
static void test(treedb_storage_on_memory_vtbl_t * v)
{
	v->ReadLock = & mem_ReadLock;
	v->WriteLock = & mem_WriteLock;
	v->Unlock = & mem_Unlock;
	v->LockType = & mem_LockType;
	v->IsOpen = & mem_IsOpen;
	v->CanWrite = & mem_CanWrite;
	v->CanExec = & mem_CanExec;
	v->Close = & mem_Close;
	v->Open = & mem_Open;
	v->OpenUnknown = & mem_OpenUnknown;
	v->Create = & mem_Create;
	v->MoreCore = & mem_MoreCore;
	v->FreeCore = & mem_FreeCore;
	v->Resizable = & mem_Resizable;
	v->PostFree = & mem_PostFree;
	v->EndianIndex = & mem_EndianIndex;
	v->EndianSwapIndex = & mem_EndianSwapIndex;
}
#endif
treedb_storage_on_memory_vtbl_t treedb_storage_on_memory_vtbl = {
	treedb_storage_vtbl_t_entries(mem_)
};
void init_treedb_storage_on_memory
( treedb_storage_on_memory_ptr_t ps, void * p
, uintptr_t size
) {
	*((treedb_storage_on_memory_vtbl_ptr_t *)(& ps->m_pVtbl))
		= & treedb_storage_on_memory_vtbl;
	ps->m_pBase = p;
	ps->m_uSize = size;
}

char * treedb_string_append(char * str, const char * more)
{
	size_t len1;
	if(str)
		len1 = strlen(str);
	else
		len1 = 0;
	size_t len2 = strlen(more);
	char * ret = (char *)realloc(str, len1 + len2 + 1);
	if(!ret)
		return ret;
	memcpy(ret + len1, more, len2 + 1);
	return ret;
}
void treedb_allocator_t_check_result
( treedb_allocator_ptr_t This
, int res
) {
	/* Don't reset any earlier error. */
	if(!res)
		return;
	/* Store the earliest error. */
	if(!This->m_errno || (This->m_errno > res)) {
		This->m_errno = res;
		if(This->m_abort_fn) {
			/* May abort. */
			fprintf(stderr, "treedb_allocator_t_check_result : Calling abort "
				"function at %p.\n", This->m_abort_fn);
			fflush(stderr);
			(*(This->m_abort_fn))(This->m_errno);
		}
		return;
	}
}
/*
treedb_allocator_ptr_t
get_allocator_on_storage_for
( treedb_storage_ptr_t pstorage
, int16_t bpo
, uint16_t flags
, uint16_t offs
) {
	treedb_const_allocator_on_storage_ptr_t ptemplate
		= treedb_get_allocator_on_storage_template_for(bpo, flags);
	if(!ptemplate)
		return 0;
	treedb_allocator_on_storage_ptr_t pallocator
		= malloc(sizeof(treedb_allocator_on_storage_t));
	if(!pallocator)
		return 0;
	* pallocator = * ptemplate;
	treedb_allocator_on_storage_Setup(pallocator);
	pallocator->m_free.m_tree_offset += offs;
	pallocator->m_used.m_tree_offset += offs;
	pallocator->m_pStorage = pstorage;
	return (treedb_allocator_ptr_t)pallocator;
}
*/

void treedb_storage_on_file_t_Setup
( treedb_storage_on_file_ptr_t ps
) {
	treedb_storage_on_file_t s = TREEDB_STORAGE_ON_FILE_INIT;
	*ps = s;
}

#ifdef __cplusplus
} // extern "C" {
#endif
